from __future__ import annotations

import dataclasses
from typing import Callable, Mapping, Sequence, Tuple, Union

from .base import Arg, FacingDef, IntOrArg, IntRelCoord, NORTH, Nbt, NbtDef, Position, RelCoord, StrOrArg, \
    _ensure_size, _in_group, _quote, _to_list, as_facing, d, de_arg, r, to_id
from .commands import Biome, Block, BlockDef, COLORS, Command, Commands, Entity, EntityDef, JsonList, JsonText, \
    SignCommand, SignCommands, SignMessage, SignMessages, SomeMappings, as_biome, as_block, as_entity, data, fill, \
    fillbiome, setblock
from .enums import Pattern

ARMORER = 'Armorer'
BUTCHER = 'Butcher'
CARTOGRAPHER = 'Cartographer'
CLERIC = 'Cleric'
FARMER = 'Farmer'
FISHERMAN = 'Fisherman'
FLETCHER = 'Fletcher'
LEATHERWORKER = 'Leatherworker'
LIBRARIAN = 'Librarian'
MASON = 'Mason'
NITWIT = 'Nitwit'
SHEPHERD = 'Shepherd'
TOOLSMITH = 'Toolsmith'
WEAPONSMITH = 'Weaponsmith'
NONE = 'None'
CHILD = 'Child'
VILLAGER_PROFESSIONS = (
    ARMORER,
    BUTCHER,
    CARTOGRAPHER,
    CLERIC,
    FARMER,
    FISHERMAN,
    FLETCHER,
    LEATHERWORKER,
    LIBRARIAN,
    MASON,
    NITWIT,
    SHEPHERD,
    TOOLSMITH,
    WEAPONSMITH,
    NONE,
)
"""Villager professions."""

DESERT = 'Desert'
JUNGLE = 'Jungle'
PLAINS = 'Plains'
SAVANNA = 'Savanna'
SNOW = 'Snow'
SWAMP = 'Swamp'
TAIGA = 'Taiga'
VILLAGER_BIOMES = (DESERT, JUNGLE, PLAINS, SAVANNA, SNOW, SWAMP, TAIGA)
"""Villager biomes / types."""


class Sign(Block):
    """A class that represents a sign. This class is a standing sign, the WallSign subclass is for wall signs."""

    waxed = False
    """Whether signs are waxed by default."""

    def __init__(self, text: SignMessages = (), /, commands: SignCommands = (), wood='oak', state: Mapping = None,
                 nbt: NbtDef = None, hanging=False, front=True):
        """
        Creates a sign object. The text and commands are passed to front().
        """
        self.hanging = hanging
        wood = to_id(wood)
        super().__init__(self._kind_name(wood), state=state, nbt=nbt)
        self.wood = wood
        if text or commands:
            self.messages(text, commands, front=front)
        self.wax(Sign.waxed)
        if nbt:
            self.merge_nbt(nbt)

    def front(self, text: SignMessages, /, commands: SignCommands = ()) -> Sign:
        """Sets the text attributes for the front of the sign."""
        self.messages(text, commands, front=True)
        return self

    def back(self, text: SignMessages, /, commands: SignCommands = ()) -> Sign:
        """Sets the text attributes for the back of the sign."""
        self.messages(text, commands, front=False)
        return self

    def messages(self, texts: SignMessages, commands: SignCommands = (), front: bool = None) -> Sign:
        """Set the text for the front, back, or both if ``front`` is None."""
        messages = self.lines_nbt(texts, commands)
        if front or front is None:
            self.merge_nbt({'front_text': messages})
        if front is False or front is None:
            self.merge_nbt({'back_text': messages})
        return self

    def glowing(self, v: bool, front: bool = None) -> Sign:
        """Set whether the text will be glowing for the front, back, or both if ``front`` is None."""
        if front or front is None:
            self.nbt.set_or_clear('front_text.has_glowing_text', v)
        if front is False or front is None:
            self.nbt.set_or_clear('back_text.has_glowing_text', v)
        return self

    def color(self, color: str = None, front: bool = None) -> Sign:
        """Set the text will color for the front, back, or both if ``front`` is None."""
        color = _in_group(COLORS, color)
        if front or front is None:
            self.nbt.set_or_clear('front_text.color', color)
        if front is False or front is None:
            self.nbt.set_or_clear('back_text.color', color)
        return self

    def wax(self, on=True):
        """Sets the sign to be waxed or not. The default is True (ignores ``Sign.waxed``)"""
        self.nbt.set_or_clear('is_waxed', on)
        return self

    @classmethod
    def lines_nbt(cls, texts: SignMessages, commands: SignCommands = ()) -> Nbt:
        """Returns the lines of NBT for sign text.
        :param texts: The sign text, as an iterable of one to four lines of text. Entries that are None will generate no
        NBT, any text will generate a line for the sign.
        :param commands: Commands for the sign, in order.
        :return: The NBT for the combination of text and commands.
        """
        texts = _ensure_size(_to_list(texts), 4)
        commands = _ensure_size(_to_list(commands), 4)
        max_count = max(len(texts), len(commands))
        if max_count > 4:
            raise ValueError(f'{max_count}: Too many values for text and/or commands')
        texts = _ensure_size(texts, 4)
        commands = _ensure_size(commands, 4)

        messages = []
        for i in range(4):
            messages.append(cls.line_nbt(texts[i], commands[i]))

        return Nbt({'messages': messages})

    def _kind_name(self, wood):
        return f'{wood}_hanging_sign' if self.hanging else f'{wood}_sign'

    @classmethod
    def line_nbt(cls, text: SignMessage = None, command: SignCommand = None) -> Nbt:
        orig_text = text
        if text is None:
            text = JsonText.text('')
        elif isinstance(text, str):
            text = JsonText.text(text)
        entry = text
        if isinstance(command, Callable):
            command = command(orig_text)
        if command:
            entry = text.click_event().run_command(command)
        return entry

    @classmethod
    def change(cls, pos: Position, messages: SignMessages = None, commands: SignCommands = None,
               front=True, start=0) -> Commands:
        messages = messages if messages else (None, None, None, None)
        commands = commands if commands else (None, None, None, None)
        cmds = []
        for f in ('front', 'back'):
            if f == 'front' and front is False:
                continue
            elif f == 'back' and front is True:
                continue
            face = f'{f}_text'
            added = 0
            for i, desc in enumerate(zip(messages, commands)):
                msg, cmd = desc
                if msg is None and cmd is None:
                    continue
                cmds.append(
                    data().modify(pos, f'{face}.messages[{i + start}]').set().value(str(cls.line_nbt(msg, cmd))))
                added += 1
            if added == 4 and start == 0:
                # If everything is being changed, this is much more efficient
                change_all = (cls.lines_nbt(messages, commands))
                to_merge = Nbt()
                if front is None:
                    to_merge['front_text'] = change_all
                    to_merge['back_text'] = change_all
                else:
                    to_merge[face] = change_all
                cmds = [data().merge(pos, to_merge)]
                break
        return cmds

    def place(self, pos: Position, facing: FacingDef, /, water=False, nbt: NbtDef = None,
              clear=True) -> Commands | Command:
        """
        Place the sign.

        :param pos: The position.
        :param facing: The direction the sign if facing. See as_facing() for useful parameters.
        :param water: Whether the sign is waterlogged.
        :param nbt: Any extra NBT for the sign.
        :param clear: Clear out the block before placing
        :return: The commands to place the sign.
        """
        self._orientation(facing)
        if water:
            self.merge_state({'waterlogged': True})
        if nbt:
            self.merge_nbt(nbt)
        if clear:
            return (
                setblock(pos, 'water' if water else 'air'),
                setblock(pos, self),
            )
        else:
            return setblock(pos, self)

    def _orientation(self, facing):
        self.merge_state({'rotation': as_facing(facing).sign_rotation})


class WallSign(Sign):
    """A class for wall signs."""

    def _kind_name(self, wood):
        return f'{wood}_wall_hanging_sign' if self.hanging else f'{wood}_wall_sign'

    def _orientation(self, facing):
        self.merge_state({'facing': as_facing(facing).name})

    def place(self, pos: Position, facing: FacingDef, /, water=False, nbt: NbtDef = None, clear=True) -> Commands:
        """When placing a wall sign, the orientations are different, but also can be found in as_facing()."""
        return super().place(pos, facing, water, nbt, clear)


class Book:
    """A class for a book."""

    def __init__(self, title: str = None, author: str = None, display_name: str = None):
        """Creates a book object."""
        self.title = title
        self.author = author
        self.display_name = display_name
        self._pages = []
        self._cur_page = JsonList()

    # Two kinds of books: Written and signed. In theory, they should hold the same kind
    # of text, but the unsigned book cannot have rich text. Hopefully in the future this _will_ be possible, so
    # this method is kept separate instead of being incorporated into the __init__ of a
    # "signed book" class that is separate from the "unsigned book" class. Or some such design.
    def sign_book(self, title: str, author: str, display_name: str = None):
        """Sign the book. An unsigned book cannot have rich text."""
        self.title = title
        self.author = author
        self.display_name = display_name

    def add(self, *txt: JsonText | str):
        """Add text to the current page of the book."""
        if self.title is None:
            raise ValueError("Cannot add Json text to unsigned book")
        for t in txt:
            if isinstance(t, str):
                t = JsonText.text(t)
            self._cur_page.append(t)

    def next_page(self):
        """Start the next page."""
        self._pages.append(self._cur_page)
        self._cur_page = JsonList()

    def as_entity(self):
        """Returns the book as an Entity object. This is useful for a ``give`` command."""
        return Entity('written_book', nbt=self.nbt())

    def as_item(self):
        """Returns the book as an Item object. This is useful for things like putting the book into a container."""
        item = Item.nbt_for('written_book')
        nbt = self.nbt()
        try:
            pages = nbt['pages']
            if pages:
                nbt['pages'] = _quote(pages)
        except KeyError:
            pass

        return Nbt({'Book': item.merge({'tag': nbt})})

    def nbt(self):
        """Returns the NBT for the book."""
        cur_page = self._cur_page
        self.next_page()
        jt = Nbt()
        jt['title'] = self.title
        jt['author'] = self.author
        if self.display_name:
            jt['display_name'] = {'Lore': self.display_name}
        jt['pages'] = [JsonList(x) for x in self._pages[:]]
        self._cur_page = cur_page
        self._pages.pop()
        return jt


class Display(Entity):
    """
    A class for the various "display" objects: text_display, item_display, and block_display. The main feature is that
    it makes transformation modifications work on the summon command; see https://bugs.mojang.com/browse/MC-259838.
    """
    INIT_TRANSFORMATION = Nbt(
        {'transformation': {'left_rotation': [0.0, 0.0, 0.0, 1.0], 'translation': [0.0, 0.0, 0.0],
                            'right_rotation': [0.0, 0.0, 0.0, 1.0], 'scale': [1.0, 1.0, 1.0]}})

    def __init__(self, id: str, nbt=None, name=None):
        super().__init__(id, nbt, name)
        # Without this, a simple change to the transform cannot be given at summon time.
        # Crazily, this is "as intended": https://bugs.mojang.com/browse/MC-259838
        self.merge_nbt(Display.INIT_TRANSFORMATION)

    def scale(self, value: float | Tuple[float, float, float]) -> Display:
        """
        Sets the scale transformation. If only given one value, it uses that for all three scale values. Otherwise,
        it must be given the three values.
        """
        if isinstance(value, float):
            value = [float(value), float(value), float(value)]
        else:
            value = tuple(float(x) for x in value)
        self.merge_nbt({'transformation': {'scale': value}})
        return self


class ItemDisplay(Display):
    """An object that represent an item_display entity."""

    def __init__(self, item: EntityDef):
        item = as_entity(item)
        nbt = Item.nbt_for(item)
        super().__init__('item_display', {'item': nbt})


def _str_values(state):
    """Convert any non-str primitive values into str, because BlockDisplay requires it (ugh)."""
    if isinstance(state, Mapping):
        for k, v in state.items():
            state[k] = _str_values(v)
        return state
    elif isinstance(state, str):
        return state
    elif isinstance(state, Sequence):
        values = []
        for v in state:
            values.append(_str_values(v))
        return values
    else:
        return Nbt.to_str(state)


class BlockDisplay(Display):
    """An object that represents a block_display entity."""

    def __init__(self, block: BlockDef):
        block = as_block(block)
        super().__init__('block_display', {'block_state': {'Name': block.id, 'Properties': _str_values(block.state)}})


class TextDisplay(Display):
    """An object that represents a text_display entity."""

    def __init__(self, text: str | JsonText | Sequence[JsonText] = None, nbt: NbtDef = None):
        """
        Creates a TextDisplay with the given text, if any. The text can be a string, a JsonText object, or a list or
        tuple of JsonText objects.
        """
        super().__init__('text_display', nbt)
        self.text(text)

    def text(self, text) -> TextDisplay:
        if isinstance(text, str):
            text = JsonText.text(text)
        elif isinstance(text, (JsonText, Sequence)):
            text = str(text).replace("'", '"')
        if text is not None:
            self.merge_nbt({'text': text})
        return self

    def _simple(self, key, value) -> TextDisplay:
        self.merge_nbt({key: de_arg(value)})
        return self


class Item(Entity):
    """An object that represents an Item."""

    def __init__(self, id: str, count: int = 1, name=None, nbt=None):
        super().__init__(id, name=name)
        self.merge_nbt({'id': id, 'Count': count})
        if nbt:
            self.merge_nbt(nbt)

    @classmethod
    def nbt_for(cls, item: BlockDef, nbt=None) -> Nbt:
        """The nbt for this item."""
        item = as_block(item)
        item_id = item.id
        if item_id.find(':') < 0:
            item_id = 'minecraft:' + item_id
        retval = Nbt({'id': item_id, 'Count': 1})
        # Filled maps are stored directly, not shunted an inner tag
        if item_id:
            if item_id == 'minecraft:filled_map':
                retval = retval.merge(item.nbt)
                if nbt:
                    retval = retval.merge(nbt)
            elif item.nbt:
                retval['tag']['BlockEntityTag'] = item.nbt
        try:
            block_state = item.state
            if block_state:
                retval['tag']['BlockStateTag'] = block_state
        except AttributeError:
            pass
        if nbt:
            return retval.merge(nbt)
        return retval


class Shield(Item):
    """A shield object."""

    def __init__(self):
        """Creates a new shield."""
        super().__init__('shield')
        self.merge_nbt({'tag': {'BlockEntityTag': {'Patterns': []}}})

    def add_pattern(self, pattern: StrOrArg | Pattern, color: IntOrArg | StrOrArg) -> Shield:
        """Add a pattern to the shield."""
        color = as_color_num(color)
        patterns = self.nbt['tag']['BlockEntityTag'].get_list('Patterns')
        if isinstance(pattern, str):
            pattern = Pattern(pattern)
        patterns.append(Nbt({'Pattern': str(pattern), 'Color': color}))
        return self

    def clear_patterns(self) -> Shield:
        """Remove all patterns from the shield."""
        self.nbt['tag']['BlockEntityTag']['Patterns'] = []
        return self


class Region:
    """Represents a region of space, and gives tools for changing items within it."""
    __slab_states = []
    __stair_states = []
    __door_states = []
    __trapdoor_states = []
    __button_states = []
    __slab_states.append(Nbt({'type': 'double'}))
    __dirs = ("north", "east", "west", "south")
    for __h in ('top', 'bottom'):
        __slab_states.append(Nbt({'type': __h}))
        for __f in __dirs:
            for __s in ('straight', "inner_left", "inner_right", "outer_left", "outer_right"):
                __stair_states.append(Nbt({'half': __h, 'facing': __f, 'shape': __s}))
            for __o in (True, False):
                __trapdoor_states.append(Nbt({'half': __h, 'facing': __f, 'open': __o}))
                for __g in ('left', 'right'):
                    __door_states.append(
                        Nbt({'half': 'upper' if __h == 'top' else 'lower', 'facing': __f, 'open': __o, 'hinge': __g}))
    for __f in __dirs:
        for __t in ('ceiling', 'floor', 'wall'):
            __button_states.append({'facing': __f, 'face': __t})
    slab_states = tuple(__slab_states)
    """The block states that a slab can have related to its placements."""
    stair_states = tuple(__stair_states)
    """The block states that stairs can have related to its placements."""
    door_states = tuple(__door_states)
    """The block states that a door can have related to its placements."""
    trapdoor_states = tuple(__trapdoor_states)
    """The block states that a trapdoor can have related to its placements."""
    button_states = tuple(__button_states)
    """The block states that a button can have related to its placements."""
    facing_states = tuple(Nbt({'facing': x}) for x in __dirs)
    """North, east, west, and south."""
    facing_all_states = tuple(Nbt({'facing': x}) for x in __dirs + ('up', 'down'))
    """North, east, west, south, up, and down."""
    rotation_states = tuple(Nbt({'rotation': x}) for x in range(16))
    """16 sign direction rotations."""
    axes_states = tuple(Nbt({'axis': x}) for x in ('x', 'y', 'z'))
    """X, y, and z. Pillars and logs use this, for example."""
    rail_states = tuple(Nbt({'shape': x}) for x in ('east_west', 'north_south') + tuple(
        f'ascending_{x}' for x in ('east', 'west', 'north', 'south')))
    """The block states that rails can have related to its placements."""
    curved_rail_states = tuple(Nbt({'shape': x}) for x in ('north_east', 'north_west', 'south_east', 'south_west'))
    """The block states that curved rails can have related to its placements."""

    def __init__(self, start: Position, end: Position):
        """Creates a new region object. Any two opposite corners will do."""
        self.start = start
        self.end = end

    def fill(self, new: BlockDef, replace: BlockDef = None) -> Command:
        """
        Returns a command that will fill the region with a block. If a second block is given, it will be the filter;
        only this kind of block will be replaced. This can, of course, be a tag.
        """
        f = fill(self.start, self.end, as_block(new))
        if replace:
            f = f.replace(replace)
        yield f

    def fillbiome(self, biome: Biome, replace: Biome = None) -> Command:
        f = fillbiome(self.start, self.end, as_biome(biome))
        if replace:
            f = f.replace(as_biome(replace))
        yield f

    def replace(self, new: BlockDef, old: BlockDef, states: SomeMappings = None,
                new_states: SomeMappings = None, shared_states: SomeMappings = None) -> Commands:
        """Returns commands that will replace one with block with another. If states are given, commands will be
        generated for each state, applied to both the fill block and the filter block. States are specified by a map,
        and can be passed as a single state, or an Iterable of them. States in new_states will only be applied to the
        new blocks. One command will be generated for each combination of the two sets of states. """
        states = _to_list(states) if states else [{}]
        new_states = _to_list(new_states) if new_states else [{}]
        if not shared_states:
            shared_states = {}
        new = as_block(new)
        old = as_block(old)

        if not states and not new_states and not shared_states:
            yield from self.fill(new, old)
        else:
            for new_added in new_states:
                n = new.clone().merge_state(new_added).merge_state(shared_states)
                o = old.clone().merge_state(shared_states)
                for s in states:
                    yield from self.fill(n.clone().merge_state(s), o.clone().merge_state(s))

    def replace_slabs(self, new: BlockDef, old: BlockDef = '#slabs', new_state: Mapping = None,
                      shared_states: SomeMappings = None) -> Commands:
        """Replaces slabs in the region using all the relevant states."""
        yield from self.replace(new, old, Region.slab_states, new_state, shared_states)

    def replace_stairs(self, new: BlockDef, old: BlockDef = '#stairs', new_state: Mapping = None,
                       shared_states: SomeMappings = None) -> Commands:
        """Replaces stairs in the region using all the relevant states."""
        yield from self.replace(new, old, Region.stair_states, new_state, shared_states)

    def replace_buttons(self, new: BlockDef, old: BlockDef = '#buttons', new_state: Mapping = None,
                        shared_states: SomeMappings = None) -> Commands:
        """Replaces buttons in the region using all the relevant states."""
        yield from self.replace(new, old, Region.button_states, new_state, shared_states)

    def replace_doors(self, new: BlockDef, old: BlockDef = '#doors', new_state: Mapping = None,
                      shared_states: SomeMappings = None) -> Commands:
        """
        Replaces doors in the region using all the relevant states.
        N.B.: Not implemented, doors cannot be replaced generically, https://bugs.mojang.com/browse/MC-192791
        """
        raise NotImplementedError('Fill does not replace doors, sorry. https://bugs.mojang.com/browse/MC-192791')
        # yield from self.replace(new, old, Region.door_states, new_state, shared_states)

    def replace_trapdoors(self, new: BlockDef, old: BlockDef = '#trapdoors', new_state: Mapping = None,
                          shared_states: SomeMappings = None) -> Commands:
        """Replaces trapdoors in the region using all the relevant states."""
        yield from self.replace(new, old, Region.trapdoor_states, new_state, shared_states)

    def replace_facing(self, new: BlockDef, old: BlockDef, new_state: Mapping = None,
                       shared_states: SomeMappings = None) -> Commands:
        """Replaces blocks in the region using all the "facing" states."""
        yield from self.replace(new, old, Region.facing_states, new_state, shared_states)

    def replace_facing_all(self, new: BlockDef, old: BlockDef, new_state: Mapping = None,
                           shared_states: SomeMappings = None) -> Commands:
        """Replaces blocks in the region using all the "all_facing" states."""
        yield from self.replace(new, old, Region.facing_all_states, new_state, shared_states)

    def replace_rotation(self, new: BlockDef, old: BlockDef, new_state: Mapping = None,
                         shared_states: SomeMappings = None) -> Commands:
        """Replaces blocks in the region using all the "rotation" states."""
        yield from self.replace(new, old, Region.rotation_states, new_state, shared_states)

    def replace_axes(self, new: BlockDef, old: BlockDef, new_state: Mapping = None,
                     shared_states: SomeMappings = None) -> Commands:
        """Replaces blocks in the region using all the "axes"" states."""
        yield from self.replace(new, old, Region.axes_states, new_state, shared_states)

    def replace_straight_rails(self, new: BlockDef, old: BlockDef = '#rails', new_state: Mapping = None,
                               shared_states: SomeMappings = None) -> Commands:
        """Replaces straight rails in the region using all the relevant states."""
        yield from self.replace(new, old, Region.rail_states, new_state, shared_states)

    def replace_curved_rails(self, new: BlockDef = "rail", old: BlockDef = '#rails',
                             new_state: Mapping = None, shared_states: SomeMappings = None) -> Commands:
        """Replaces curved rails in the region using all the relevant states."""
        yield from self.replace(new, old, Region.curved_rail_states, new_state, shared_states)


class Offset:
    """
    This provides a tool for offsetting relative coordinates. This allows you to write code placed in a way that may
    be more convenient, such as if a command block is hidden in a convenient place, but wants to operate relative to
    a different base location. Given an initial offset of a given number of coordinates, coordinates generated
    through the object's r() and d() methods will be adjusted by that location. The values passed to r() and () must
    be the same length as the initial offset.
    """

    CoordsIn = Union[float, RelCoord]
    # This is so complicated because some functions take specific-length tuples, so we want to declare that we produce
    # them to make the type checker happier.
    CoordsOut = Union[
        RelCoord, Tuple[RelCoord, RelCoord], Tuple[IntRelCoord, IntRelCoord], Tuple[RelCoord, RelCoord, RelCoord],
        float, Tuple[float, float], Tuple[int, int], Tuple[float, float, float], Tuple[int, int, int],
        Tuple[RelCoord, ...]]

    def __init__(self, *position: float):
        """Creates an offsetting object with the given values."""
        if len(position) == 0:
            raise ValueError(f'Must have at least one value in offset')
        self.position: tuple[float] = position

    def r(self, *values: CoordsIn) -> CoordsOut:
        """ Returns the result of base.r() with the input, with each return value added to this object's offset. """
        return self._rel_coord(r, *values)

    def d(self, *values: CoordsIn) -> CoordsOut:
        """ Returns the result of base.d() with the input, with each return value added to this object's offset. """
        return self._rel_coord(d, *values)

    def p(self, *values: CoordsIn) -> CoordsOut:
        """ Returns the result of offsetting the input, with each return value added to this object's coordinates. """
        return tuple(sum(i) for i in zip(values, self.position))

    def _rel_coord(self, f, *values: CoordsIn) -> RelCoord | Tuple[RelCoord, ...]:
        if len(values) != len(self.position):
            raise ValueError(f'{len(values)} != position length ({len(self.position)})')
        vec = []
        exemplar = f(0)
        for v in values:
            if isinstance(v, (float, int)):
                vec.append(f(v))
            elif v.prefix != exemplar.prefix:
                raise ValueError(f'{f}: incompatible RelCoord type')
            else:
                vec.append(v)
        vals = RelCoord.add(vec, self.position)
        if len(vals) == 1:
            return vals[0]
        return vals


class ItemFrame(Entity):
    """A class for item frames."""

    def __init__(self, facing: int | str, *, glowing: bool = False, nbt: NbtDef = None, name: str = None):
        """Creates an ItemFrame object facing in the given direction. See as_facing() for useful values."""
        nbt = Nbt.as_nbt(nbt) if nbt else Nbt({})
        nbt = nbt.merge({'Facing': as_facing(facing).number, 'Fixed': True})
        super().__init__('glow_item_frame' if glowing else 'item_frame', nbt=nbt, name=name)

    def item(self, item: BlockDef | Entity) -> ItemFrame:
        """Sets the item that is in the frame."""
        block = as_block(item)
        self.merge_nbt({'Item': Item.nbt_for(block)})
        return self

    def fixed(self, value: bool) -> ItemFrame:
        self.nbt.set_or_clear('Fixed', value)
        return self

    def named(self, name: BlockDef = None) -> ItemFrame:
        """Sets the name displayed for the item in the frame."""
        block = as_block(name)
        if block is None:
            try:
                del self.nbt['Item']['tag']['display']['Name']
            except KeyError:
                pass  # Must not be there already, ignore the error
        else:
            if 'Item' not in self.nbt:
                self.item(block)
            nbt = self.nbt
            nbt['Item']['tag']['display']['Name'] = JsonText.text(block.name)
        return self


@dataclasses.dataclass
class Trade:
    """Represents a single trade a villager can make."""
    max_uses = 12
    uses = 0
    xp = 1
    buy: tuple[tuple[BlockDef, int]] | tuple[tuple[BlockDef, int], tuple[BlockDef, int]]
    sell: tuple[BlockDef, int]
    reward_exp = True

    def __init__(self, buy1: BlockDef | tuple[BlockDef, int], thing1: BlockDef | tuple[BlockDef, int],
                 thing2: BlockDef | tuple[BlockDef, int] = None, /, max_uses=None, uses=0, xp=1, reward_exp=True):
        """
        Creates a Trade object. The first block or item is the price. If thing2 is present, then it is the block or item
        being sold, and thing1 is the second part of the price. Otherwise, thing1 is the block or item being sold.
        """
        if thing2:
            self.buy = (_to_def(buy1), _to_def(thing1))
            self.sell = _to_def(thing2)
        else:
            self.buy = (_to_def(buy1),)
            self.sell = _to_def(thing1)
        self.max_uses = max_uses
        self.uses = uses
        self.xp = xp
        self.reward_exp = reward_exp

    def nbt(self):
        """Returns the nbt for this trade."""
        values = Nbt({
            'buy': {'id': self.buy[0][0], 'Count': self.buy[0][1]},
            'sell': {'id': self.sell[0], 'Count': self.sell[1]},
            'rewardExp': self.reward_exp
        })
        if len(self.buy) > 1:
            values['buyB'] = {'id': self.buy[1][0], 'Count': self.buy[1][1]}
        values.set_or_clear('maxUses', self.max_uses)
        return values


def _to_def(block) -> tuple[Block, int]:
    if isinstance(block, tuple):
        return as_block(block[0]).id, block[1]
    return as_block(block).id, 1


class Villager(Entity):
    """Convenience class for a villager or zombie villager. This presents simpler mechanisms for profession,
    biome, experience, levels, and trades."""
    level_xp = {
        'Novice': range(0, 10),
        'Apprentice': range(10, 70),
        'Journeyman': range(70, 150),
        'Expert': range(150, 250),
        'Master': range(250, 2147483647),
    }
    """The range of experience for each level."""

    def __init__(self, profession: str = 'Unemployed', biome: str = 'Plains', nbt: NbtDef = None, /, name=None,
                 zombie: bool = False):
        """Creates a villager."""
        super().__init__('zombie_villager' if zombie else 'villager', nbt=nbt, name=name)
        self.zombie = zombie
        self.profession(profession)
        self.biome(biome)
        self.xp(0)
        self._trades: list[Trade] = []

    def xp(self, xp: int) -> Villager:
        """Sets the villager's experience."""
        self.nbt['VillagerData']['xp'] = xp
        self.merge_nbt({'VillagerData': {'xp': xp, 'level': self.level}})
        return self

    @property
    def level(self) -> int:
        """The villager's level as a number."""
        i, _ = self._lookup_level()
        return i

    @property
    def level_name(self) -> str:
        """The villager's level as a name."""
        _, n = self._lookup_level()
        return n

    def _lookup_level(self):
        try:
            xp = self.nbt['VillagerData']['xp']
        except KeyError:
            xp = 0
        for i, (n, r) in enumerate(Villager.level_xp.items()):
            if xp in r:
                return i, n
        raise ValueError(f'{xp}: Invalid experience value')

    def profession(self, profession: str) -> Villager:
        """Sets the villager's profession. The profession can also be 'child' for non-zombie villagers."""
        profession = profession.title()
        if profession == 'Child':
            if self.zombie:
                raise ValueError('Child: Invalid zombie villager profession')
            profession = NONE
            self.merge_nbt({'Age': -2147483648})
        self.merge_nbt({'VillagerData': {'profession': _in_group(VILLAGER_PROFESSIONS, profession).lower()}})
        return self

    def biome(self, biome: str) -> Villager:
        """Sets the villager's biome."""
        biome = biome.title()
        self.merge_nbt({'VillagerData': {'type': _in_group(VILLAGER_BIOMES, biome).lower()}})
        return self

    type = biome
    """Alias for the ``biome`` method, because these two terms are used interchangeably."""

    def add_trade(self, *trades: Trade | NbtDef) -> Villager:
        """Add trades to the villager's list."""
        recipes = self.nbt['Offers'].get_list('Recipes')
        for t in trades:
            if isinstance(t, Trade):
                if len(t.buy) not in range(1, 3):
                    raise ValueError(f'{len(t.buy)}: Invalid buy length (must be 1 or 2)')
                t = t.nbt()
            recipes.append(t)
        return self

    def inventory(self, *items: BlockDef | tuple[BlockDef, int]) -> Villager:
        """Sets the villager's inventory."""
        inventory = self.nbt.get_list('Inventory')
        for i in items:
            if not isinstance(i, tuple):
                i = (i, 1)
            item_nbt = Item.nbt_for(as_block(i[0]))
            item_nbt['Count'] = i[1]
            inventory.append(Nbt(item_nbt))
        return self


@dataclasses.dataclass
class PaintingInfo:
    """
    Class that holds detailed information on each painting, available from a Painting object's ``info`` field, or the
    Painting.INFO map.
    """
    id: str
    name: str
    size: Tuple[int, int]
    added: str
    desc: str
    original_artist: str = 'None'
    based_on: str = None
    source: str = None
    artist: str = 'Kristoffer Zetterstrand'
    used: bool = True


#         'stage': PaintingInfo('stage', 'The Stage is Set', (2, 2), 'Indev 20100223 / Alpha v1.1.1',
#                               """Scenery from Space Quest I, with the character Graham from the video game series
#                               King's Quest.""",
#                               source='Space Quest I / King\'s Quest'),

class Painting(Entity):
    """
    Represents a painting. This understands its different "facing" numbering, and the PaintingInfo objects
    """
    INFO = {
        'alban': PaintingInfo('alban', 'Albanian', (1, 1), 'Indev 20100223',
                              """A man wearing a fez next to a house and a bush. As the name of the painting
                              suggests, it may be a landscape in Albania"""),
        'aztec': PaintingInfo('aztec', 'de_aztec', (1, 1), 'Indev 20100223',
                              """Free-look perspective of the map 'de_aztec' from the video game 
                              <i>Counter-Strike.</i>""",
                              based_on='de_aztec', source='Counter-Strike'),
        'aztec2': PaintingInfo('aztec', 'de_aztec', (1, 1), 'Indev 20100223',
                               """Free-look perspective of the map 'de_aztec' from the video game 
                               <i>Counter-Strike.</i>""",
                               based_on='de_aztec', source='Counter-Strike'),
        'bomb': PaintingInfo('bomb', 'Target Successfully Bombed', (1, 1), 'Indev 20100223',
                             """The map 'de_dust2' from the video game <i>Counter-Strike,</i> named “target 
                             successfully bombed" in reference to the game.""",
                             based_on='de_dust2', source='Counter-Strike'),
        'kebab': PaintingInfo('kebab', 'Kebab med tre pepperoni', (1, 1), 'Indev 20100223',
                              """A kebab with three green chili peppers."""),
        'plant': PaintingInfo('plant', 'Paradisträd', (1, 1), 'Indev 20100223',
                              """Still life of two plants in pots. "Paradisträd" is Swedish for "money tree",
                              which is a common name for the depicted species in Scandinavia."""),
        'wasteland': PaintingInfo('wasteland', 'Wasteland', (1, 1), 'Indev 20100223',
                                  """Some wastelands; a small animal (presumably a rabbit) is sitting on the window
                                  ledge."""),
        'courbet': PaintingInfo('courbet', 'Bonjour Monsieur Courbet', (2, 1), 'Indev 20100223',
                                """Two hikers with pointy beards seemingly greeting each other. This painting is
                                based on Gustave Courbet's painting <i>The Meeting</i> or <i>"Bonjour,
                                Monsieur Courbet"</i>.""",
                                based_on='Bonjour Monsieur Courbet', original_artist='Gustave Courbet'),
        'pool': PaintingInfo('pool', 'The Pool', (2, 1), 'Indev 20100223',
                             """Some men and women skinny-dipping in a pool over a cube of sorts. Also there is an
                             old man resting in the lower-right edge."""),
        'sea': PaintingInfo('sea', 'Seaside', (2, 1), 'Indev 20100223 / Alpha v1.1.1',
                            """Mountains and a lake, with a small photo of a mountain and a bright-colored plant on
                            the window ledge."""),
        'creebet': PaintingInfo("creebet", 'Creebet', (2, 1), 'Alpha v1.1.1',
                                """Mountains and a lake, with a small photo of a mountain and a creeper looking at
                                the viewer through a window."""),
        'sunset': PaintingInfo('sunset', 'Sunset Dense', (2, 1), 'Indev 20100223', """Mountains at sunset."""),
        'graham': PaintingInfo('graham', 'Graham', (1, 2), 'Alpha v1.1.1',
                               """King Graham, the player character in the video game series <i>King's Quest<i>.""",
                               based_on='Still Life with Quince, Cabbage, Melon, and Cucumber',
                               original_artist='Juan Sánchez Cotán', source='King\'s Quest'),
        'wanderer': PaintingInfo('wanderer', 'Wanderer', (1, 2), 'Indev 20100223',
                                 """A version of Caspar David Friedrich's famous painting Wanderer above the Sea of 
                                 Fog.""",
                                 based_on='Wanderer above the Sea of Fog', original_artist='Caspar David Friedrich'),
        'bust': PaintingInfo('bust', 'Bust', (2, 2), 'Indev 20100223',
                             """Bust of Marcus Aurelius surrounded by fire."""),
        'match': PaintingInfo('match', 'Match', (2, 2), 'Indev 20100223',
                              """A hand holding a match, causing fire on a white cubic gas fireplace."""),
        'skull_and_roses': PaintingInfo('skull_and_roses', 'Skull and Roses', (2, 2), 'Indev 20100223',
                                        """A skeleton at night with red flowers in the foreground."""),
        'stage': PaintingInfo('stage', 'The Stage is Set', (2, 2), 'Indev 20100223 / Alpha v1.1.1',
                              """Scenery from Space Quest I, with the character Graham from the video game series
                              King's Quest.""",
                              source='Space Quest I / King\'s Quest'),
        'void': PaintingInfo('void', 'The void', (2, 2), 'Indev 20100223',
                             """An angel praying into a void with fire below."""),
        'wither': PaintingInfo('wither', 'Wither', (2, 2), 'Java Edition 1.4.2',
                               """The creation of a wither. This is the only painting not based on a real painting.""",
                               artist='Jens Bergensten'),
        'fighters': PaintingInfo('fighters', 'Fighters', (4, 2), 'Indev 20100223',
                                 """Two men poised to fight. Paper versions of fighters from the game
                                 <i>International Karate +</i>.""",
                                 source='International Karate +'),
        'donkey_kong': PaintingInfo('donkey_kong', 'Kong', (4, 3), 'Alpha v1.1.1',
                                    """Level 100 from the arcade game <i>Donkey Kong.</i>""", source='Donkey Kong'),
        'skeleton': PaintingInfo('skeleton', 'Mortal Coil', (4, 3), 'Alpha v1.1.1',
                                 """Bruno Martinez from the adventure game <i>Grim Fandango.</i>""",
                                 source='Grim Fandango'),
        'burning_skull': PaintingInfo('burning_skull', 'Skull on Fire', (4, 4), 'Beta 1.2_01 / Beta 1.3',
                                      """A Skull on fire; in the background there is a moon in a clear night sky.
                                      Based on a <i>Minecraft</i> screenshot, with grass blocks and a 3D skull added
                                      on top.""",
                                      source='Minecraft'),
        'pigscene': PaintingInfo('pigscene', 'Pigscene', (4, 4), 'Alpha v1.1.1',
                                 """A girl pointing to a pig on a canvas.""", based_on='The Artist\'s Studio',
                                 original_artist='Jacob van Oost'),
        'pointer': PaintingInfo('pointer', 'Pointer', (4, 4), 'Indev 20100223',
                                """The main character of the game <i>International Karate +</i> in a fighting stance
                                touching a large hand. It can be interpreted as a play on Michelangelo's famous
                                painting "The Creation of Adam".""",
                                source='International Karate +'),
        'earth': PaintingInfo('earth', 'Earth', (2, 2), '22w16a', """The classical element Earth""", used=False),
        'fire': PaintingInfo('fire', 'Fire', (2, 2), '22w16a', """The classical element Fire""", used=False),
        'water': PaintingInfo('water', 'Water', (2, 2), '22w16a', """The classical element Water""", used=False),
        'wind': PaintingInfo('wind', 'Wind', (2, 2), '22w16a', """The classical element Wind""", used=False),
    }
    """PaintingInfo objects for paintings"""

    def __init__(self, variant: str):
        variant = to_id(variant)
        if variant not in self.INFO:
            raise ValueError(f'{variant}: Unknown painting variant')
        self.variant = variant
        super().__init__('painting', nbt={'variant': variant})

    @property
    def info(self):
        return Painting.INFO[self.variant]

    def summon(self, pos: Position, nbt: NbtDef = None, facing: FacingDef = NORTH, lower_left=False) -> str:
        """
        Overrides to add ``ll`` parameter which, if true, uses the lower-left corner consistently for all sizes of
        paintings.
        """
        nbt, facing = self._summon_clean(nbt, facing)
        if facing:
            nbt['facing'] = facing.h_number
        del nbt['Facing']
        del nbt['Rotation']
        nbt['variant'] = self.variant
        if lower_left:
            movement = facing.turn(90)
            x, y = self.info.size
            pos = _to_list(pos)
            if x > 2:
                adj = x - 3
                pos[0] += adj * movement.dx
                pos[2] += adj * movement.dz
            if y > 2:
                pos[1] += 1
        return super().summon(pos, nbt)


def as_color(color: IntOrArg | StrOrArg | None) -> str | None:
    """Checks if the argument is a valid color name, or None.

    "Valid" means one of the 16 known colors, such as those used for wool. These are stored in the
    ``COLORS`` array.

    An Arg is also valid.

    :param color: The ((probable) color name.
    :return: The color name, in lower case.
    """
    if isinstance(color, Arg):
        return str(color)
    if color is None:
        return None
    color_num = as_color_num(color)
    return COLORS[color_num]


def as_color_num(color: IntOrArg | StrOrArg | None) -> int | str | None:
    """Checks if the argument is a valid color number specification, or None.

    "Valid" means an int, or a string that names a known color from which a color number can be inferred.
    Color numbers range from 0 to 15. (See as_color() for a documentation on color names.)

    An Arg is also valid.

    :param color:
    :return:
    """
    if isinstance(color, Arg):
        return str(color)
    if color is None:
        return None
    if isinstance(color, str):
        color_num = COLORS.index(to_id(color))
        if color_num < 0:
            raise ValueError(f'{color}: Unknown color')
        return color_num
    if color not in range(len(COLORS)):
        raise ValueError(f'{color}: Unknown color')
    return color



import unittest

from parameterized import parameterized

from pynecraft.base import Arg, COLORS, Coord, EAST, IntRelCoord, NORTH, Nbt, RED, ROTATION_0, ROTATION_180, \
    ROTATION_270, ROTATION_90, RelCoord, SOUTH, TimeSpec, WEST, _bool, _ensure_size, _float, _in_group, _int_or_float, \
    _not_ify, _quote, _strip_namespace, _strip_not, _to_list, _to_tuple, as_angle, as_column, as_duration, as_facing, \
    as_name, \
    as_names, as_nbt_key, as_nbt_path, as_pitch, as_range, as_resource, as_resource_path, as_resources, as_yaw, d, days, \
    de_arg, r, \
    rotate_facing, seconds, settings, string, ticks, to_id
from pynecraft.commands import setblock


def spaceify(s, use_spaces):
    return s.replace(',', ', ').replace(':', ': ') if use_spaces else s


class TestBase(unittest.TestCase):
    def setUp(self) -> None:
        self.orig_spaces = Nbt.use_spaces

    def tearDown(self) -> None:
        Nbt.use_spaces = self.orig_spaces

    def test_quote(self):
        self.assertEqual('foo', _quote('foo'))
        self.assertEqual('"f b"', _quote('f b'))
        self.assertEqual('\'"f b"\'', _quote('"f b"'))
        self.assertEqual('"\'f b\'"', _quote("'f b'"))
        self.assertEqual('"true"', _quote("true"))
        self.assertEqual('"false"', _quote("false"))
        self.assertEqual(1, _quote(1))
        self.assertEqual(0.1, _quote(.1))
        self.assertEqual(1.0, _quote(1.))
        self.assertEqual(1.1, _quote(1.1))
        self.assertEqual('"1."', _quote("1."))
        self.assertEqual('".1"', _quote(".1"))
        self.assertEqual('"1.1"', _quote("1.1"))
        self.assertEqual('t1', _quote("t1"))

    def test_to_list(self):
        self.assertListEqual([], _to_list(()))
        self.assertListEqual([], _to_list([]))
        self.assertListEqual([None], _to_list(None))
        self.assertListEqual([1], _to_list((1,)))
        self.assertListEqual([1], _to_list([1, ]))
        self.assertListEqual(['a'], _to_list('a'))
        self.assertListEqual([1], _to_list(1))
        self.assertListEqual([0, 1, 2], _to_list(range(3)))

    def test_to_tuple(self):
        self.assertTupleEqual((), _to_tuple(()))
        self.assertTupleEqual((), _to_tuple([]))
        self.assertTupleEqual((None,), _to_tuple(None))
        self.assertTupleEqual((1,), _to_tuple((1,)))
        self.assertTupleEqual((1,), _to_tuple([1, ]))
        self.assertTupleEqual(('a',), _to_tuple('a'))
        self.assertTupleEqual((1,), _to_tuple(1))
        self.assertTupleEqual((0, 1, 2), _to_tuple(range(3)))

    def test_strip_namespace(self):
        self.assertEqual('foo', _strip_namespace('m:foo'))
        self.assertEqual('foo', _strip_namespace('foo'))
        self.assertEqual('$(b)', _strip_namespace(Arg('b')))
        self.assertEqual('a$(b)', _strip_namespace('a$(b)'))

    def test_strip_not(self):
        self.assertEqual('foo', _strip_not('!foo'))
        self.assertEqual('foo', _strip_not('foo'))
        self.assertEqual('$(f)', _strip_not(Arg('f')))
        self.assertEqual('b$(f)', _strip_not('b$(f)'))
        self.assertEqual('b$(f)', _strip_not('!b$(f)'))

    def test_bool(self):
        self.assertIsNone(_bool(None))
        self.assertEqual('true', _bool(True))
        self.assertEqual('$(b)', _bool(Arg('b')))

    def test_float(self):
        self.assertEqual('1', _float(1))
        self.assertEqual('1.0', _float(1.0))
        self.assertEqual('1.1', _float(1.1))
        self.assertEqual('1.123', _float(1.12345))
        self.assertEqual('$(f)', _bool(Arg('f')))

    def test_not_ify(self):
        self.assertEqual('!foo', _not_ify('foo'))
        self.assertTupleEqual(('!foo', '!bar', '!baz'), _not_ify(('foo', 'bar', 'baz')))

    def test_ensure_size(self):
        self.assertListEqual([None, None], _ensure_size([], 2))
        self.assertListEqual([None, None], _ensure_size([None], 2))
        self.assertListEqual([None, None], _ensure_size([None, None], 2))
        with self.assertRaises(ValueError):
            _ensure_size([None, None, None], 2)

    def test_as_nbt_key(self):
        self.assertEqual('key', as_nbt_key('key'))
        self.assertEqual('v$(k)', as_nbt_key('v$(k)'))
        self.assertEqual('$(k)', as_nbt_key(Arg('k')))

    def test_de_arg(self):
        self.assertEqual('x', de_arg('x'))
        self.assertEqual('$(x)', de_arg('$(x)'))
        self.assertEqual('$(x)', de_arg(Arg('x')))
        self.assertEqual(('x', 'y'), de_arg(('x', 'y')))
        self.assertEqual(('x', '$(y)'), de_arg(('x', Arg('y'))))
        self.assertEqual(('x', ('$(y)', '$(z)')), de_arg(('x', (Arg('y'), Arg('z')))))

    def test_as_nbt_path(self):
        self.assertEqual('a.b.c', as_nbt_path('a.b.c'))
        self.assertEqual('$(a)', as_nbt_path(Arg('a')))
        self.assertEqual('a.$(b).c', as_nbt_path('a.$(b).c'))
        self.assertEqual('a.b[-1]', as_nbt_path('a.b[-1]'))
        with self.assertRaises(ValueError):
            as_nbt_path('')
        with self.assertRaises(ValueError):
            as_nbt_path('a.b%c')

    def test_as_resource(self):
        self.assertIsNone(as_resource(None))
        self.assertEqual('a', as_resource('a'))
        self.assertEqual('!a', as_resource('!a', allow_not=True))
        self.assertEqual('m:a', as_resource('m:a'))
        self.assertEqual('!m:a', as_resource('!m:a', allow_not=True))
        self.assertEqual('!v$(k)', as_resource('!v$(k)', allow_not=True))
        self.assertEqual('v$(k)', as_resource('v$(k)', allow_not=True))
        self.assertEqual('$(k)', as_resource('$(k)', allow_not=True))
        self.assertEqual('$(k)', as_resource(Arg('k'), allow_not=True))
        with self.assertRaises(ValueError):
            as_resource('!a')
        with self.assertRaises(ValueError):
            as_resource('m:a', allow_namespace=False)

    def test_as_resources(self):
        self.assertTupleEqual((), as_resources())
        self.assertTupleEqual(('a',), as_resources('a'))
        self.assertTupleEqual(('a', '!b'), as_resources('a', '!b', allow_not=True))
        self.assertTupleEqual(('v$(a)', '!v$(b)'), as_resources('v$(a)', '!v$(b)', allow_not=True))
        self.assertTupleEqual(('$(a)', '$(b)'), as_resources(Arg('a'), Arg('b'), allow_not=True))
        with self.assertRaises(ValueError):
            as_resources('!a')
        with self.assertRaises(ValueError):
            as_resources('a', '!b')

    def test_as_resource_path(self):
        self.assertIsNone(as_resource_path(None))
        self.assertEqual('a', as_resource_path('a'))
        self.assertEqual('/a', as_resource_path('/a'))
        self.assertEqual('/a/b/c', as_resource_path('/a/b/c'))
        self.assertEqual('!/a/b/c', as_resource_path('!/a/b/c', allow_not=True))
        self.assertEqual('!v$(k)', as_resource_path('!v$(k)', allow_not=True))
        self.assertEqual('v$(k)', as_resource_path('v$(k)', allow_not=True))
        self.assertEqual('$(k)', as_resource_path('$(k)', allow_not=True))
        self.assertEqual('$(k)', as_resource_path(Arg('k'), allow_not=True))
        with self.assertRaises(ValueError):
            as_resource_path('')
        with self.assertRaises(ValueError):
            as_resource_path('!/a/b/c')
        with self.assertRaises(ValueError):
            as_resource_path('//a/b/c')

    def test_as_name(self):
        self.assertIsNone(as_name(None))
        self.assertEqual('a', as_name('a'))
        self.assertEqual('!a', as_name('!a', allow_not=True))
        self.assertEqual('v$(a)', as_name('v$(a)'))
        self.assertEqual('$(a)', as_name(Arg('a')))
        with self.assertRaises(ValueError):
            as_name('!a')
        with self.assertRaises(ValueError):
            as_name('a%b')

    def test_as_names(self):
        self.assertTupleEqual((), as_names())
        self.assertTupleEqual(('a',), as_names('a'))
        self.assertTupleEqual(('!a',), as_names('!a', allow_not=True))
        self.assertTupleEqual(('!a', 'b', 'c'), as_names('!a', 'b', 'c', allow_not=True))
        with self.assertRaises(ValueError):
            as_names('!a')

    def test_as_column(self):
        self.assertEqual((1, 2), as_column((1, 2)))
        self.assertEqual(r(1, 2), as_column(r(1, 2)))
        self.assertEqual((1, r(2)), as_column((1, r(2))))
        self.assertEqual(('v$(a)',), as_column('v$(a)'))
        self.assertEqual(('$(a)',), as_column(Arg('a')))
        with self.assertRaises(ValueError):
            as_column((1,))
        with self.assertRaises(ValueError):
            as_column((1, 2, 3))

    def test_as_angle(self):
        self.assertEqual(17, as_angle(17))
        self.assertEqual(17.3, as_angle(17.3))
        self.assertEqual(r(17.3), as_angle(r(17.3)))
        self.assertEqual('v$(a)', as_angle('v$(a)'))
        self.assertEqual('$(a)', as_angle(Arg('a')))
        with self.assertRaises(ValueError):
            as_angle(d(17.3))

    def test_as_yaw(self):
        self.assertIsNone(as_yaw(None))
        self.assertEqual(17.3, as_yaw(17.3))
        self.assertEqual(90, as_yaw(WEST))
        self.assertEqual('-$(a).1', as_yaw('-$(a).1'))
        self.assertEqual('$(a)', as_yaw(Arg('a')))
        with self.assertRaises(ValueError):
            as_yaw(181)
        with self.assertRaises(ValueError):
            as_yaw(-181)
        with self.assertRaises(KeyError):
            as_yaw('v$(k)')

    def test_as_pitch(self):
        self.assertIsNone(as_pitch(None))
        self.assertEqual(17.3, as_pitch(17.3))
        self.assertEqual('-$(a).1', as_pitch('-$(a).1'))
        self.assertEqual('$(a)', as_pitch(Arg('a')))
        with self.assertRaises(ValueError):
            as_pitch(181)
        with self.assertRaises(ValueError):
            as_pitch(-181)
        with self.assertRaises(TypeError):
            as_pitch('v$(k)')
        with self.assertRaises(TypeError):
            as_pitch(WEST)

    def test_in_group(self):
        self.assertIsNone(_in_group(COLORS, None))
        self.assertEqual(RED, _in_group(COLORS, RED))
        self.assertEqual('v$(a)', _in_group(COLORS, 'v$(a)'))
        self.assertEqual('$(a)', _in_group(COLORS, Arg('a')))
        with self.assertRaises(ValueError):
            _in_group(COLORS, None, allow_none=False)
        with self.assertRaises(ValueError):
            _in_group(COLORS, 'ecru')

    def test_as_facing(self):
        self.assertEqual(90, as_facing(WEST).yaw)
        self.assertEqual(0, as_facing(WEST).pitch)
        self.assertEqual('ese', as_facing(13).name)

    def test_rotated_facing(self):
        self.assertEqual((0, 0, -3), as_facing(NORTH).scale(3))
        self.assertEqual(as_facing(NORTH), rotate_facing(NORTH, ROTATION_0))
        self.assertEqual(as_facing(NORTH), rotate_facing(EAST, ROTATION_270))
        self.assertEqual(as_facing(NORTH), rotate_facing(SOUTH, ROTATION_180))
        self.assertEqual(as_facing(NORTH), rotate_facing(WEST, ROTATION_90))

    def test_as_duration(self):
        self.assertIsNone(as_duration(None))
        self.assertEqual(TimeSpec('15s'), as_duration(TimeSpec('15s')))
        self.assertEqual(TimeSpec(15), as_duration(15))
        self.assertEqual('v$(k)', as_duration('v$(k)'))
        self.assertEqual('$(k)', as_duration(Arg('k')))

    def test_as_range(self):
        self.assertEqual('17', as_range(17))
        self.assertEqual('28.3', as_range(28.3))
        self.assertEqual('5.3..8.4', as_range((5.3, 8.4)))
        self.assertEqual('5.3..', as_range((5.3, None)))
        self.assertEqual('..8.4', as_range((None, 8.4)))
        self.assertEqual('$(k)', as_range('$(k)'))
        self.assertEqual('+$(k).1', as_range('+$(k).1'))
        self.assertEqual('$(k)', as_range(Arg('k')))
        self.assertEqual('$(k)..-2.$(v)3', as_range((Arg('k'), '-2.$(v)3')))
        with self.assertRaises(ValueError):
            as_range((6, 3))
        with self.assertRaises(ValueError):
            as_range(False)
        with self.assertRaises(ValueError):
            as_range('v$(k)')
        with self.assertRaises(ValueError):
            as_range(('v$(k)', 'q$(z)'))

    def test_string(self):
        self.assertEqual('', string(''))
        self.assertEqual('3', string(3))
        self.assertEqual('(1, 2)', string((1, 2)))
        self.assertEqual('(1, ~2, ^3)', string((1, r(2), d(3))))

    def test_to_id(self):
        self.assertEqual('foo', to_id('foo'))
        self.assertEqual('foo_bar', to_id('Foo Bar'))
        self.assertEqual('v$(k)', to_id('v$(k)'))
        self.assertEqual('$(k)', to_id(Arg('k')))

    def test_nbt(self):
        self.assertEqual({'key': 1}, Nbt().merge(Nbt(key=1)))
        self.assertEqual({'key': 1}, Nbt(key=1).merge(None))
        self.assertEqual({'key': 1, 'key2': 2}, Nbt(key=1).merge(Nbt(key2=2)))
        self.assertEqual({'key': 2}, Nbt(key=1).merge(Nbt(key=2)))
        self.assertEqual({'key': 1, 'key2': 3}, Nbt(key=1, key2=2).merge(Nbt(key2=3)))
        self.assertEqual({'key': (2, 4, 6)}, Nbt(key=(1, 3, 5)).merge(Nbt(key=(2, 4, 6))))
        self.assertEqual({'key': '➝'}, Nbt(key='➝'), 'Non-ascii unicode should be unchanged')
        simple_nbt = Nbt(One=2)
        self.assertIs(simple_nbt, Nbt.as_nbt(simple_nbt))
        self.assertIsNot(simple_nbt, simple_nbt.clone())
        self.assertEqual(simple_nbt, simple_nbt.clone())
        self.assertEqual(Nbt(), Nbt()['One'])

    def test_set_or_clear(self):
        self.assertEqual({'key': 12}, Nbt().set_or_clear('key', 12))
        self.assertEqual({'key': 0}, Nbt().set_or_clear('key', 0))
        self.assertEqual({}, Nbt(key=12).set_or_clear('key', None))
        self.assertEqual({'key': 'foo'}, Nbt().set_or_clear('key', 'foo'))
        self.assertEqual({}, Nbt(key=12).set_or_clear('key', None))
        self.assertEqual({'o1': {'o2': {'o3': {'key': True}}}}, Nbt().set_or_clear('o1.o2.o3.key', True))
        self.assertEqual({'o1': {'o2': {'o3': {}}}},
                         Nbt({'o1': {'o2': {'o3': {'key': True}}}}).set_or_clear('o1.o2.o3.key', False))

    def test_nbt_str(self):
        self.assertEqual('{}', Nbt.to_str({}))
        self.assertEqual('{}', Nbt.to_str(Nbt({})))
        self.assertEqual('[]', Nbt.to_str([]))
        self.assertEqual(Nbt(sub=Nbt()), Nbt.as_nbt({'sub': {}}))
        self.assertEqual('$(a)', Nbt.to_str(Arg('a')))

    def test_nbt_regularize(self):
        self.assertEqual('{key: [1, 2]}', str(Nbt(key=[1, 2])))
        self.assertEqual('{key: [1.0f, 2.0f]}', str(Nbt(key=[1, 2.0])))
        self.assertEqual('{key: [1.0f, 2.0f]}', str(Nbt(key=[1.0, 2.0])))

    def test_nbt_forced_types(self):
        for key in ('Rotation', 'LeftArm', 'RightArm', 'LeftLeg', 'RightLeg', 'Head', 'Body'):
            self.assertEqual(f'{{{key}: [1f, 2f]}}', str(Nbt({key: [1, 2]})))
        self.assertEqual('{Motion: [1d, 2d]}', str(Nbt(Motion=[1, 2])))

    def test_nbt_get_list(self):
        self.assertEqual([], Nbt().get_list('key'))
        self.assertEqual([1, 2], Nbt(key=[1, 2]).get_list('key'))
        self.assertEqual(Arg('a'), Nbt({'list': Arg('a')}).get_list('list'))

    def test_nbt_get_nbt(self):
        self.assertEqual(Nbt(), Nbt().get_nbt('key'))
        self.assertEqual(Nbt(good=True), Nbt(key={'good': True}).get_nbt('key'))

    @parameterized.expand(((True,), (False,),))
    def test_nbt_format(self, spaces):
        Nbt.use_spaces = spaces
        self.assertEqual(spaceify('{}', spaces), str(Nbt()))
        self.assertEqual(spaceify('{key:true}', spaces), str(Nbt(key=True)))
        self.assertEqual(spaceify('{key:1}', spaces), str(Nbt(key=1)))
        # Could be either order
        self.assertIn(str(Nbt(key='value', key2=2)), (
            spaceify('{key:value,key2:2}', spaces), spaceify('{key2:2,key:value}', spaces)))
        self.assertEqual(spaceify('{key:"val ue"}', spaces), str(Nbt(key='val ue')))
        self.assertEqual(spaceify('{key:{num:17}}', spaces), str(Nbt(key={'num': 17})))
        self.assertEqual(spaceify('{key:{nums:[1,2,3]}}', spaces), str(Nbt(key={'nums': [1, 2, 3]})))

        with self.assertRaises(KeyError):
            Nbt({'k-ey': 13})

    @parameterized.expand(((True,), (False,),))
    def test_nbt_array(self, spaces):
        Nbt.use_spaces = spaces
        self.assertEqual(spaceify('[B;123]', spaces), str(Nbt.TypedArray('b', (123,))))
        self.assertEqual(spaceify('[B;123,4,5,6]', spaces), str(Nbt.TypedArray('B', (123, 4, 5, 6))))
        self.assertEqual(spaceify('[I;123]', spaces), str(Nbt.TypedArray('i', (123,))))
        self.assertEqual(spaceify('[I;123,4,5,6]', spaces), str(Nbt.TypedArray('I', (123, 4, 5, 6))))
        self.assertEqual(spaceify('[L;123]', spaces), str(Nbt.TypedArray('l', (123,))))
        self.assertEqual(spaceify('[L;123,4,5,6]', spaces), str(Nbt.TypedArray('L', (123, 4, 5, 6))))

        with self.assertRaises(ValueError):
            Nbt.TypedArray('d', ())

    def test_precision(self):
        orig = settings.float_precision
        try:
            settings.float_precision = 1
            self.assertEqual('setblock 1.1 2.2 5.6 air', str(setblock((1.111, 2.222, 5.555), 'air')))
            settings.float_precision = 3
            self.assertEqual('setblock 1.111 2.222 5.555 air', str(setblock((1.111, 2.222, 5.555), 'air')))

            with self.assertRaises(ValueError):
                settings.float_precision = 0

        finally:
            settings.float_precision = orig

    def test_rel_coord(self):
        self.assertEqual(IntRelCoord('~', 3), r(3))
        self.assertEqual(RelCoord('~', 3.0), r(3.0))
        self.assertEqual(RelCoord('~', 3.1), r(3.1))
        self.assertEqual(RelCoord('~', Arg('v')), r(Arg('v')))
        self.assertEqual(r(5.5), r(2) + r(3.5))
        self.assertEqual(r(5.5), r(2) + 3.5)
        self.assertEqual(r(3.5), r(5.5) - r(2))
        self.assertEqual(r(3.5), r(5.5) - 2)
        self.assertEqual(r(7.2), r(4) * r(1.8))
        self.assertEqual(r(7.2), r(4) * 1.8)
        self.assertEqual(r(5.5), r(11) / r(2))
        self.assertEqual(r(5.5), r(11) / 2)
        self.assertEqual(r(5), r(11) // r(2))
        self.assertEqual(r(5), r(11) // 2)
        self.assertEqual(r(9), r(3) ** 2)
        self.assertEqual(r(4), abs(r(-4)))
        self.assertEqual(r(4), abs(r(4)))
        self.assertTrue(r(3) < r(3.1))
        self.assertFalse(r(3) > r(3.1))
        self.assertFalse(r(-3.5) > -r(3.5))
        self.assertEqual(r(-3.5), -r(3.5))
        self.assertEqual(r(3.5), +r(3.5))
        self.assertEqual(r(4, 6), RelCoord.add(r(1, 2), r(3, 4)))
        self.assertEqual(r(-2, -2), RelCoord.sub(r(1, 2), r(3, 4)))

        with self.assertRaises(AssertionError):
            r(3.5) + d(3.5)
        with self.assertRaises(AssertionError):
            r(Arg('v')) + d(3.5)

    def test_rel_coord_merge(self):
        def add(v1: Coord, v2: Coord):
            return v1 + v2

        self.assertEqual(None, RelCoord.merge(add, None, None))
        self.assertEqual((1, 2, 3), RelCoord.merge(add, (1, 2, 3), None))
        self.assertEqual((1, 2, 3), RelCoord.merge(add, None, (1, 2, 3)))
        self.assertEqual((2, 4, 6), RelCoord.merge(add, (1, 2, 3), (1, 2, 3)))
        self.assertEqual(r(1, 2, 3), RelCoord.merge(add, r(1, 2, 3), None))
        self.assertEqual(r(1, 2, 3), RelCoord.merge(add, None, r(1, 2, 3)))
        self.assertEqual(r(2, 4, 6), RelCoord.merge(add, r(1, 2, 3), r(1, 2, 3)))
        with self.assertRaises(TypeError):
            RelCoord.merge(add, r(Arg('a'), 2), r(1, 2))

    def test_time_spec(self):
        self.assertEqual(0, TimeSpec(0).ticks)
        self.assertEqual(1, TimeSpec(0.9).ticks)
        self.assertEqual(1, TimeSpec(1).ticks)
        self.assertEqual(2, TimeSpec(1.2).ticks)
        self.assertEqual(1, TimeSpec(20).seconds)
        self.assertEqual(1, TimeSpec(24_000).days)
        self.assertEqual(20, TimeSpec('1s').ticks)
        self.assertEqual(24_000, TimeSpec('1d').ticks)
        self.assertEqual(TimeSpec('15d'), days(15))
        self.assertEqual(TimeSpec('15s'), seconds(15))
        self.assertEqual(TimeSpec('15t'), ticks(15))
        self.assertEqual(TimeSpec(15), ticks(15))
        self.assertEqual('15d', str(TimeSpec('15d')))
        self.assertEqual('15s', str(TimeSpec('15s')))
        self.assertEqual('15', str(TimeSpec('15t')))
        self.assertEqual('15', str(TimeSpec('15')))
        with self.assertRaises(ValueError):
            TimeSpec('5q')

    def test_int_or_float(self):
        self.assertTrue(isinstance(_int_or_float(1), int))
        self.assertTrue(isinstance(_int_or_float(1.0), int))
        self.assertTrue(isinstance(_int_or_float(1.5), float))

    def test_arg(self):
        self.assertEqual('$(foo)', str(Arg('foo')))
        self.assertEqual(Arg('a'), Arg('a'))
        self.assertEqual(Arg('a'), '$(a)')
        self.assertNotEqual(Arg('a'), 'a')
        self.assertIsNotNone(Arg('a'))
        self.assertEqual(hash(Arg('a')), hash(Arg('a')))
        self.assertNotEqual(hash(Arg('a')), hash(Arg('b')))
        with self.assertRaises(ValueError):
            Arg('')



#!/usr/bin/env python

"""

Rebuilds pynecraft.all*.txt using current data. This is inherently a bit crufty and sensitive, because it reads data
from a wiki, which may change at any time. If I knew of a better way to get this info, I'd use it.

"""

from __future__ import annotations

import re
import sys
from abc import ABC, abstractmethod
from pathlib import Path

import requests
from bs4 import BeautifulSoup

cwd = Path(sys.path[0])


class Fetcher(ABC):
    def __init__(self, which: str, url: str):
        self.which = which
        self.url = url

    @abstractmethod
    def get_start(self, page):
        pass

    @abstractmethod
    def is_end(self, elem) -> bool:
        pass

    @abstractmethod
    def get_id(self, raw_id: str, raw_desc: str) -> tuple[str, str]:
        pass

    def added(self, things: list[str]):
        return []

    def desc_elem(self):
        return 'li'

    def fetch(self):
        html = requests.get(self.url).text
        page = BeautifulSoup(html, 'html.parser')
        start = self.get_start(page)
        things = []
        for elem in start.find_next_siblings():
            if self.is_end(elem):
                break
            for li in elem.find_all(self.desc_elem()):
                raw_text = li.text.replace(u'\u200c', '')  # Discard the zero-width non-joiners
                m = re.search(r'\[(.*) only]', raw_text, re.IGNORECASE)
                if m and not ('Java' in m.group(1) or 'JE' in m.group(1)):
                    continue

                raw_desc = raw_text.strip()
                raw_id = re.sub(r'\s+', ' ', raw_desc.replace("'", " "))
                id, desc = self.get_id(raw_id, raw_desc)
                if id:
                    row = desc if id == desc else f'{desc} / {id}'
                    things.append(row)

        for to_add in self.added(things):
            assert to_add not in things
            things.append(to_add)

        with open(cwd / f'../all_{self.which}.txt', 'w') as fp:
            fp.write('\n'.join(sorted(things)))
            fp.write('\n')


class BlockFetcher(Fetcher):
    def __init__(self):
        super().__init__('blocks', 'https://minecraft.fandom.com/wiki/Block#List_of_blocks')

    def get_start(self, page):
        return page.find('h2', string='List of blocks')

    def is_end(self, elem):
        return elem.name == 'h2' or 'Technical blocks' in elem.text

    def get_id(self, raw_id, raw_desc):
        if 'Ominous' in raw_id or 'Torchflower Crop' in raw_id or 'Pitcher Crop' in raw_id:
            # This is not really a block at all.
            return None, None

        # this appears once randomly, and it shouldn't
        if ' (block)' in raw_id:
            raw_id = re.sub(' \(block\)', '', raw_id)
        id = raw_id
        desc = raw_desc
        if 'upcoming' in id:
            if not re.search(r'\bJ\.*E\.*', id):
                return None, None
            id = re.sub(r'\s*\[*upcoming.*', '', id)
            desc = re.sub(r'\s*\[*upcoming.*', ' [x]', desc)
        if 'Lapis' in id and id != 'Lapis Lazuli':
            id = id.replace('Lapis Lazuli', 'Lapis')
        elif 'Bale' in id:
            id = id.replace('Bale', 'Block')
        elif 'Redstone' in id:
            id = re.sub(r'Redstone (Repeater|Comparator)', r'\1', id)
        elif id == 'Monster Spawner':
            id = 'Spawner'
        elif id == 'Grass':
            id = 'Short Grass'
        if 'Block' in id or 'Crops' in id:
            id = re.sub(r'Block of (.*)', r'\1 Block', id)
            id = re.sub(r'(Jigsaw|Light|Smooth Quartz|Wheat) (Block|Crops)', r'\1', id)
        id = re.sub(r'^(Vine)s', r'\1', id)
        id = re.sub(r'Bamboo Shoot', 'Bamboo Sapling', id)
        return id, desc

    def added(self, _):
        # These should be there, but aren't
        return ['Air', 'Cave Air']


class ItemFetcher(Fetcher):
    must_give = [
        'Knowledge Book',
    ]
    operator_menu = [
        'Barrier',
        'Command Block',
        'Chain Command Block',
        'Repeating Command Block',
        'Minecart with Command Block',
        'Debug Stick',
        'Jigsaw',
        'Light',
        'Structure Block',
        'Structure Void',
    ]

    id_replace = {'Redstone Dust': 'Redstone', 'Book and Quill': 'Writable Book', 'Empty Map': 'Map',
                  'Steak': 'Cooked Beef', 'Turtle Shell': 'Turtle Helmet', 'Disc Fragment': 'Disc Fragment 5',
                  'Nether Quartz': 'Quartz', 'Slimeball': 'Slime Ball'}

    def __init__(self):
        super().__init__('items', 'https://minecraft.fandom.com/wiki/Item?so=search#List_of_items')

    def get_start(self, page):
        return page.find('h2', string='List of items')

    def is_end(self, elem):
        return elem.name == 'h2' or 'Education Edition' in elem.text

    def get_id(self, raw_id, raw_desc):
        #  This is in the list as a way to say "any potion", it's not an item.
        if 'Potions' in raw_id:
            return None, None

        id = raw_id
        desc = raw_desc

        if 'Music Disc' in raw_desc:
            id = desc = raw_desc.replace('(', '').replace(')', '')
        elif 'Banner Pattern ' in raw_desc:
            m = re.fullmatch(r'Banner Pattern \(([^ ]+)( .*)?.*\)', desc)
            id = f'{m.group(1)} Banner Pattern'.replace('Snout', 'Piglin').replace('Thing', 'Mojang')
            desc = m.group(1)
        elif ' Trim' in id:
            id = id.replace(' Trim', ' Trim Smithing Template')
        elif 'Netherite Upgrade' in id:
            id = 'Netherite Upgrade Smithing Template'
        id = re.sub(r'\s*[([].*', '', id)
        desc = re.sub(r'\s*[([].*', '', desc)

        if 'Explorer Map' in id or 'Command Block Minecart' in id:
            # These show up twice
            return None, None
        elif id in self.id_replace:
            id = self.id_replace[id]
        elif 'Bottle o\'' in desc:
            id = 'Experience Bottle'
        elif 'Banner Pattern ' in id:
            m = re.fullmatch(r'Banner Pattern \(([^ ]+)( .*)?.*\)', id)
            id = f'{m.group(1)} Banner Pattern'
            desc = m.group(1)
        elif re.search('(Boat|Raft) with Chest', id):
            id = re.sub(r'(.*) (Boat|Raft) with Chest', r'\1 Chest \2', id)
        elif id == 'Monster Spawner':
            id = 'spawner'
        elif ' with ' in id:
            id = re.sub(r'(.*) with (.*)', r'\2 \1', id)
        elif 'Redstone Dust' in id:
            id = 'Redstone'
        elif 'Book and Quill' in id:
            id = 'Writable Book'
        elif ' Cap' in id:
            id = id.replace('Cap', 'Helmet')
        elif ' Pants' in id:
            id = id.replace('Pants', 'Leggings')
        elif ' Tunic' in id:
            id = id.replace('Tunic', 'Chestplate')
        elif 'Bucket of' in id or 'Eye of' in id:
            id = re.sub(r'(Bucket|Eye) of (.*)', r'\2 \1', id)
        elif ' s ' in id:  # was 's before replacement above
            id = id.replace(" s ", ' ')
        elif 'Raw ' in id:
            m = re.fullmatch('Raw (Copper|Iron|Gold)', id)
            if not m:
                id = id[4:]
        elif id == 'Scute':
            id = 'Turtle Scute'

        return id, desc

    def added(self, things):
        # In this case, some names do show up in the list, and which ones might change.
        return filter(lambda s: s not in things, ItemFetcher.must_give)


class MobFetcher(Fetcher):
    def __init__(self):
        super().__init__('mobs', 'https://minecraft.fandom.com/wiki/Mob?so=search#List_of_mobs')

    def get_start(self, page):
        return page.find('h2', string='List of mobs')

    def is_end(self, elem):
        return elem.name == 'h2' or 'Unused mobs' in elem.text or 'Upcoming' in elem.text

    def desc_elem(self):
        return 'td'

    def get_id(self, raw_id: str, raw_desc: str):
        id = re.sub(r'\s*\(.*', '', raw_id)
        desc = re.sub(r'\s*\(.*', '', raw_desc)
        # Not really mobs, you can't summon them.
        if 'Jockey' in desc or 'Horseman' in desc:
            return None, None
        return id, desc

    def added(self, things: list[str]):
        if 'Breeze' not in things:
            return ['Breeze']
        return super().added(things)


if __name__ == '__main__':
    BlockFetcher().fetch()
    ItemFetcher().fetch()
    MobFetcher().fetch()



"""Functionality related to functions."""

from __future__ import annotations

import math
import os
import shutil
from re import Pattern
from typing import Any, MutableMapping

from .base import _JsonEncoder, _in_group, _to_list, _to_tuple
from .commands import *

BLOCKS = 'blocks'
FLUIDS = 'fluids'
ITEMS = 'items'
ENTITIES = 'entity_types'
EVENTS = 'game_events'
TAG_SETS = [BLOCKS, FLUIDS, ITEMS, ENTITIES, EVENTS]

ADVENTURE = 'adventure'
END = 'end'
HUSBANDRY = 'husbandry'
NETHER = 'nether'
RECIPES = 'recipes'
STORY = 'story'
ADVANCEMENT_SETS = [ADVENTURE, END, HUSBANDRY, NETHER, RECIPES, STORY]

NOISE_SETTINGS = 'noise_settings'
BIOME = 'biome'
CONFIGURED_CARVER = 'configured_carver'
CONFIGURED_SURFACE_BUILDER = 'configured_surface_builder'
CONFIGURED_FEATURE = 'configured_feature'
CONFIGURED_STRUCTURE_FEATURE = 'configured_structure_feature'
STRUCTURE_SET = 'structure_set'
TEMPLATE_POOL = 'template_pool'
PROCESSOR_LIST = 'processor_list'
WORLDGEN_SETS = [NOISE_SETTINGS, BIOME, CONFIGURED_CARVER, CONFIGURED_SURFACE_BUILDER, CONFIGURED_FEATURE,
                 CONFIGURED_STRUCTURE_FEATURE, STRUCTURE_SET, TEMPLATE_POOL, PROCESSOR_LIST]


def text_lines(*orig: any) -> Iterable[str]:
    """Converts a number of commands and lines into a sequence of single lines, each terminated by newlines."""
    result = []
    for cmd in lines(orig):
        text = str(cmd)
        if len(text) > 0 or not text.endswith('\n'):
            text += '\n'
        result.append(text)
    return result


def as_function_name(name: str):
    """
    Checks if the name is a valid function name.

    :param name: The (probable) function name.
    :return: the input value.
    """
    m = re.fullmatch(r'(\w+:)?(\w+/)?(\w+)', name)
    if not m:
        raise ValueError(f'{name}: Invalid function name')
    return name


def _write_json(values: Mapping, path: Path):
    with open(path, 'w') as fp:
        json.dump(values, fp, cls=_JsonEncoder, indent=4, ensure_ascii=False)
        fp.write('\n')


class Function:
    """A class that represents a function."""
    SUFFIX = '.mcfunction'

    _LOAD_INFO = '# __internal_load: '

    def __init__(self, name: str, base_name: str = None):
        self.name = as_function_name(name)
        self.base_name = base_name if base_name else name
        self._add_to = []
        self.parent = None

    @property
    def full_name(self) -> str:
        """The full name of the function, including datapack name and path."""
        if self.parent:
            parent_name = self.parent.full_name
            if not parent_name.endswith(':'):
                parent_name += '/'
            return parent_name + self.name
        return self.name

    @property
    def pack(self):
        """The pack this function is in."""
        ancestor = self.parent
        while ancestor is not None and ancestor.pack is None:
            ancestor = ancestor.parent
        if ancestor and ancestor.pack:
            return ancestor.pack
        return None

    def __str__(self):
        return self.name + ':' + str('\n'.join(self.commands()) + '\n')

    @classmethod
    def load(cls, path: Path | str) -> Function:
        if isinstance(path, str):
            path = Path(path)
        if path.suffix != Function.SUFFIX:
            if path.suffix:
                raise ValueError(f'{path}: Invalid function suffix')
            path = path.with_suffix(Function.SUFFIX)
        with open(path) as fp:
            lines = fp.readlines()
            if len(lines) == 0 or not lines[-1].startswith(Function._LOAD_INFO):
                return Function._load(lines, path, {})
            spec = lines.pop()
            load_info = json.loads(spec[len(Function._LOAD_INFO):])
            if load_info['type'] == 'Function':
                return Function._load(lines, path, load_info)
            elif load_info['type'] == 'Loop':
                # noinspection PyProtectedMember
                return Loop._load(lines, path, load_info)
            else:
                raise SyntaxError(f'{load_info["type"]}: Invalid LOAD_INFO type')

    @classmethod
    def _load(cls, lines, path, load_info):
        func = Function(load_info.get('name', path.stem), load_info.get('base_name', None))
        func.add(lines)
        return func

    def _load_info(self) -> dict:
        return {'type': self.__class__.__name__, 'name': self.name, 'base_name': self.base_name}

    def save(self, path: Path | str = None) -> Path:
        """Saves the function to a file.

        If the path is not given, it is derived from this function's full path name, using the Minecraft file layout.
        If the path is a directory, the function is written into a file with the name of the function followed by
        ``.mcfunction``. If the path is to a file, it must end in ``.mcfunction`` or have no suffix, in which case
        ``.mcfunction`` will be added.
        """
        path = self._path_for(path)
        path.parent.mkdir(exist_ok=True, parents=True)
        with open(path, 'w') as out:
            out.writelines(text_lines(self.commands()))
            out.write(Function._LOAD_INFO)
            out.write(json.dumps(self._load_info()))
            out.write('\n')
        return path

    def _path_for(self, path):
        if path is None:
            path = Path(self.name.split(':')[-1])
        path = path if isinstance(path, Path) else Path(path)
        if path.is_dir():
            path = path / self.name
        if path.suffix:
            if path.suffix != Function.SUFFIX:
                raise ValueError(f'{path.suffix}: Suffix must be absent or "{Function.SUFFIX}"')
        else:
            path = path.with_suffix(Function.SUFFIX)
        return path

    def add(self: T, *cmds: [Command | str]) -> T:
        """Adds commands to the function.

        You can provide a list of strings or Command objects, or any un-flat iterables of them. They will be flattened
        into a set of strings for the function's commands.
        """
        cmds = lines(cmds)
        self._add_to.extend(cmds)
        return self

    def commands(self) -> list[str]:
        """Returns the commands in the function."""
        return self._add_to


class Loop(Function):
    """
    A loop function. This does the following:

    * Creates a score that is incremented each time the function is called.
    * Executes a series of commands for each iteration of the loop, only running those for the specific iteration value.
    * Generates commands for a "cur" function that will execute the code without iteration, which is useful when the
      commands are conditional on something else that may have changed.

    The loop body is provided with the loop() function. It is given a function that returns the commands for a given
    iteration of the loop, and then makes those commands run only when the score has that iteration's value. The
    commands can be as similar or dissimilar for each iteration as your loop function makes it. The other parameter
    to loop() is the list of things that are the values for each step. This could be a simple range() command,
    or a list of names, or whatever.

    As an example, you could have the following:

    ```
    Loop(('foo', 'funcs')).loop(lambda step: say(f'{step.i}: {step.elem}, ('Red', 'Green', 'Blue'))
    ```

    The lambda will be invoked three times. The ``step`` argument defines the data for the specific iteration; ``i``
    is the iteration number, and ``elem`` is the value for from the list for that iteration. This will produce a
    function that looks something like this:

    ```
    # ...increment Score('foo', 'funcs') for the loop...
    execute if score foo funcs matches 0 run say 0: Red
    execute if score foo funcs matches 1 run say 1: Green
    execute if score foo funcs matches 2 run say 2: Blue
    ```

    You can have a function that is much more complicated than this lambda, one that might (for example) use
    ``step.i`` to craft each loop iteration. It's up to you.

    The loop has four parts: ``setup`` which handles the loop increments; ``before`` which is run before the loop body;
    ``body`` which is the loop body; and ``after`` which is run after the rest. These parts are available as
    fields, though ``setup`` and ``loop`` are read-only, and created  automatically when you invoke loop().
    You can call loop() at most once unless you
    force it with the 'replace' parameter.

    Commands that are added using add() will be executed before or after the loop every time the function is invoked,
    depending on whether you call add() before or after loop().
    """

    iterate_at = 10
    """The threshold for breaking the loop out into iterations. A negative value disables this feature. 
    
    It is easy to see that there is one comparison per line of the iterations. When iterations have a lot of lines, 
    this results in long scripts with lots of tests. This value says that when any iteration is over this number of 
    lines, each iteration will be broken out into a separate function, which will be invoked with one test. So 
    instead of (say) 12 iterations of 10 lines each giving one function with 120 tests, there will be 12 tests, 
    each of which invokes a separate 10 line function. This give more functions, but fewer tests. 
    
    Benchmarking is hard with Minecraft on this point. This is our (rather random) guess at the tradeoff for this value.
    """

    @dataclasses.dataclass
    class Step:
        """Data about a specific step in the iteration."""
        i: int
        stage: int
        count: int
        elem: any
        last: any
        loop: Loop

    class TestControls:
        """Controls that are useful to simplify behavior for testing Loop. You should not otherwise use this code."""

        @staticmethod
        def set_prefix_override(func):
            Loop._prefix_override = func

        @staticmethod
        def set_setup_override(lines: tuple[str]):
            Loop._setup_override = lines

    _prefix_override = None
    _setup_override = None

    @classmethod
    def test_controls(cls):
        return Loop.TestControls()

    def __init__(self, score: ScoreName, name=None, base_name=None):
        """
        Creates a Loop function that uses the specified score for its iteration counter.

        :param score: The score to use for the iteration number.
        :param name: The loop name, if not give, it will be the 'target' part of the score.
        :param base_name: The base name for the function. It can be useful to name the function 'foo_main' to run the
        function on the "main" clock, for example. If so, the base name would typically be 'foo'.
        """
        score = as_score(score)
        if not name:
            if isinstance(score.target, (str, User)):
                name = str(score.target)
            else:
                raise ValueError(f'{score.target}: Not a valid name, and no name specified')
        super().__init__(name, base_name)
        self.score = score
        self.looped = False
        self.to_incr = Score('_to_incr', score.objective)
        self.max_score = Score(self.score.target, f'{self.score.objective}_max')
        self.setup = ()
        self.adjuster = ()
        self.before = []
        self.body = ()
        self.after = []
        self._iterations = []
        self._add_to = self.before

    @classmethod
    def _load(cls, lines, path, load_info):
        loop = Loop(load_info['score'], load_info.get('name', None), load_info.get('base_name', None))
        loop.looped = load_info['looped']

        lines = [x.rstrip() for x in lines]
        loop.setup = tuple(cls._pop_lines(lines, load_info['setup_len']))
        adjuster_len = load_info['adjuster_len']
        if not adjuster_len:
            loop.adjuster = ()
        else:
            adjuster_pos = -1 - adjuster_len
            loop.adjuster = loop.setup[adjuster_pos:-1]
            loop.setup = loop.setup[0:adjuster_pos] + loop.setup[-1:-1]
        loop.before = cls._pop_lines(lines, load_info['before_len'])
        loop.body = tuple(cls._pop_lines(lines, load_info['body_len']))
        loop.after = lines
        loop._iterations = load_info['iterations']

        if load_info.get('as_iterations', True) and loop.body:
            new_body = []
            for i, it in enumerate(loop._iterations):
                iter_suffix = loop._iter_suffix(i)
                prefix = str(loop._prefix_for(i)) + ' '
                name = loop.name + iter_suffix
                func = Function.load(path.parent / name)
                lines = func.commands()
                assert len(lines) == it
                for line in lines:
                    new_body.append(prefix + line)
            loop.body = tuple(new_body)

        loop._add_to = loop.before if not loop.looped else loop.after
        return loop

    @classmethod
    def _pop_lines(cls, lines, pos):
        return [lines.pop(0) for _ in range(pos)]

    def _load_info(self):
        info = super()._load_info()
        info.update({
            'looped': self.looped,
            'score': (str(self.score.target), self.score.objective),
            'before_len': len(self.before),
            'adjuster_len': len(self.adjuster),
            'setup_len': len(self.setup),
            'body_len': len(self.body),
            'iterations': self._iterations,
            'as_iterations': self._as_iteration(),
        })
        return info

    def save(self, path: Path | str = None) -> Path:
        as_iteration = self._as_iteration()
        orig_body = self.body
        try:
            # Clear out old iteration files if they exist
            full_path = self._path_for(path)
            for f in full_path.parent.glob(full_path.name + '__[0-9]*' + Function.SUFFIX):
                f.unlink()

            if as_iteration:
                new_body = []
                pos = 0
                for i in range(0, len(self._iterations)):
                    iter_suffix = self._iter_suffix(i)
                    prefix = str(self._prefix_for(i)) + ' '
                    new_body.append(prefix + str(function(self.full_name + iter_suffix)))
                    iter_name = self.name + iter_suffix
                    func = Function(iter_name)
                    func.add(x[len(prefix):] for x in self.body[pos:pos + self._iterations[i]])
                    pos += self._iterations[i]
                    func.save(full_path.parent / iter_name)
                self.body = new_body
            return super().save(path)
        finally:
            if as_iteration:
                self.body = orig_body

    def _as_iteration(self):
        return len(self._iterations) > 0 and 0 <= Loop.iterate_at <= max(self._iterations)

    def _iter_suffix(self, i):
        width = int(math.log(len(self._iterations), 10)) + 1
        return f'__{i:0{width}}'

    def commands(self) -> list[str]:
        return list(self.setup) + self.before + list(self.body) + self.after

    def _setup_for(self, loop_size: int):
        if Loop._setup_override:
            return _to_tuple(Loop._setup_override())
        setup = [
            execute().unless().score(self.score).matches((0, None)).run(str(function(
                f'{self.score.target}_init'))),
            self.max_score.set(loop_size),
            execute().if_().score(self.to_incr).matches((1, None)).run(literal(self.score.add(1)))]
        setup.extend(self.adjuster)
        setup.append(self.score.operation(MOD, self.max_score))
        return tuple(setup)

    def _prefix_for(self, i):
        if Loop._prefix_override:
            return Loop._prefix_override(i)
        return execute().if_().score(self.score).matches(i).run('')

    def loop(self,
             body_func: Callable[[Step], Union[Commands, Command, str, Iterable[Union[Commands, Command, str]]]] | None,
             items: Iterable[Any],
             bounce: object = False, replace: object = False) -> Loop:
        """
        Define the loop itself.

        :param body_func: The function to call for each iteration of the loop. This returns the same kind of values
                          that add() will take.
        :param items: The items that will be iterated over. Can be a simple range or any other iterable.
        :param bounce: If True, the loop will "bounce" between the list values. A loop with range(0, 4) and bounce True
                       will go through the values (0, 1, 2, 3, 2, 1) instead of (0, 1, 2, 3).
        :param replace: If True, will replace the loops contents; otherwise you will get an AssertionError.
        :return: the Loop object.
        """
        if not replace:
            assert not self.looped, 'loop() invoked more than once'

        items = _to_list(items)
        last = items[-1] if len(items) > 0 else None
        count = len(items)
        stages = tuple(range(count))
        if bounce and count > 2:
            items = items + list(reversed(items[1:-1]))
            stages = stages + tuple(reversed(stages[1:-1]))

        self.setup = self._setup_for(len(items))
        self._add_to = []

        self._iterations = []
        if body_func:
            for i, elem in enumerate(items):
                # noinspection PyArgumentList
                once = lines(body_func(Loop.Step(i, stages[i], count, elem, last, self)))
                self._iterations.append(len(once))
                prefix = str(self._prefix_for(i)) + ' '
                for line in lines(once):
                    self.add(str(prefix) + line)

        self.body = tuple(self._add_to)
        self._add_to = self.after

        self.looped = True
        return self

    def cur(self) -> Commands:
        """Return commands for a "cur" function that will run the function without incrementing the score."""
        return lines(
            self.to_incr.set(0),
            function(self.full_name),
            self.to_incr.set(1),
        )

    def adjust(self, *adjuster: [Command | str]) -> Loop:
        """
        Execute a command or commands after the loop value is incremented but before it is constrained to the max
        value. This can be used, for example, to skip a value in the middle if it is not compatible with another
        loop's value.
        """
        self.adjuster = tuple(lines(adjuster))
        return self


LATEST_PACK_VERSION = 18


class DataPack:
    """
    A datapack. This maintains a datapack directory. If it is being stored in a world save, it knows how to find its
    location therein.
    """

    def __init__(self, name: str, format_version: int = LATEST_PACK_VERSION, /,
                 mcmeta: Mapping = None):
        self._name = name
        self.function_set = FunctionSet('functions', self)
        self._json = {}
        self._mcmeta = {'pack': {'pack_format': format_version, 'description': JsonText.text(name)}}
        self._description = None
        if mcmeta:
            self._mcmeta.update(mcmeta)

    @classmethod
    def path_for(cls, path: Path | str, name: str = None) -> tuple[Path, str]:
        path = path if isinstance(path, Path) else Path(path)
        if not name:
            name = path.name
        else:
            datapack_path = path / 'datapacks'
            if datapack_path.exists():
                path = datapack_path / name
                path.mkdir(exist_ok=True)
            elif path.name != name:
                path /= name
        return path, name

    @classmethod
    def load(cls, path: Path | str, name: str = None) -> DataPack:
        """Loads a datapack from the files in the path.

        If not specified, the name of the pack will be the final path component."""
        path, name = cls.path_for(path, name)
        if not path.exists():
            raise ValueError(f'{path}: No such directory')
        pack = DataPack(name)
        pack.function_set = FunctionSet.load(path / pack._data_dir(), pack, 'functions')
        pack._json = pack._load_dict(pack._data_dir())
        with open(path / 'pack.mcmeta') as fp:
            pack._mcmeta = json.load(fp)
        return pack

    def _load_dict(self, path: Path) -> dict:
        contents = {}
        for entry in path.glob('*'):
            path = Path(entry)
            if path.is_file():
                if path.suffix != '.json':
                    print(f'Warning: {path} ignored: not a .json file')
                    continue
                with open(path) as fp:
                    contents[path.stem] = json.load(fp)
            else:
                if entry.stem == 'functions' and path.stem == self.name:
                    continue
                contents[path.stem + '/'] = self._load_dict(path)
        return contents

    def save(self, path: Path | str):
        """Saves all the files in the datapack. This first removes the path, if it exists, to clean out any old
        files. It then writes all the new files, as well as the ``pack.mcmeta`` file and a warning ``README`` file.
        """
        path, _ = self.path_for(path, self._name)
        if path.exists():
            shutil.rmtree(path)
        path.mkdir(parents=True)
        self._cleanup()
        self.function_set.save(path / self._data_dir())
        self._save_dict(self._json, path / self._data_dir())
        _write_json(self._mcmeta, path / 'pack.mcmeta')
        with open(path / 'README', 'w') as fp:
            fp.write("Files in this tree were auto-generated using pynecraft. Hand modifications will be lost!")

    def _cleanup(self) -> None:
        if 'tags/' in self._json:
            self._cleanup_values(self._json['tags/'])

    def _cleanup_values(self, values: MutableMapping) -> None:
        for k, v in values.items():
            if isinstance(v, MutableMapping):
                if 'values' not in v:
                    self._cleanup_values(v)
            elif isinstance(v, Iterable):
                values[k] = {'values': _to_list(v)}

    def _save_dict(self, d: dict, path: Path):
        for k, v in d.items():
            if k.endswith('/'):
                subdir = path / k[:-1]
                subdir.mkdir(exist_ok=True, parents=True)
                self._save_dict(v, subdir)
            else:
                _write_json(v, (path / k).with_suffix('.json'))

    @property
    def name(self):
        """The data pack name."""
        return self._name

    @property
    def description(self) -> JsonText | None:
        """The data pack description, stored in the ``pack.mcmeta`` file."""
        try:
            return self._mcmeta['pack']['description']
        except KeyError:
            return None

    @description.setter
    def description(self, text: JsonDef):
        if isinstance(text, str):
            text = JsonText.text(text)
        self._mcmeta['pack']['description'] = text

    @property
    def filter(self) -> list[Mapping[str, str]] | None:
        """The filter listing, stored in the ``pack.mcmeta`` file."""
        try:
            return self._mcmeta['pack']['filter']
        except KeyError:
            return None

    @filter.setter
    def filter(self, value: Iterable[tuple[str, str]]):
        self._mcmeta['pack']['filter'] = _to_list(value)

    @property
    def mcmeta(self) -> dict:
        """The dict for the ``pack.mcmeta`` file."""
        return self._mcmeta

    @property
    def functions(self):
        """The FunctionSet for the top-level functions in this data pack."""
        return self.function_set.functions

    def _data_dir(self):
        return Path('data') / self.name

    def add_filter(self, namespace: str = None, pattern: str | Pattern = None):
        """Adds a new filter for the pack."""
        f = {}
        if namespace is not None:
            f['namespace'] = namespace
        if pattern is not None:
            if isinstance(pattern, str):
                re.compile(pattern)
            else:
                pattern = pattern.pattern
            f['path'] = str(pattern)
        # noinspection PyTypeChecker
        self._mcmeta['pack'].setdefault('filter', []).append(f)

    def tags(self, tag_set: str) -> dict:
        """
        Returns defined tags. You can add to this dict to add tags to the datapack. As a convenience, if a tag's
        value is a list (or tuple), it is replaced with {'values': *tag_value*}, so you don't have to remember that.
        So saying ``tags(BLOCKS)['hard'] = ['stone', 'slate']`` is the same as saying  ``tags(BLOCKS)['hard'] = {
        'values': ['stone', 'slate']}``. And it's easier.
        """
        return self._get_json('tags', TAG_SETS, tag_set)

    def advancements(self, advancement_set: str) -> dict[str, dict]:
        """Returns defined advancements. You can add to this dict to add advancements to the datapack. Keys that end in
        '/' define directories. Keys underneath them (which do not end in '/') define files, whose contents are the JSON
        defined by the dict such keys map to."""
        return self._get_json('advancements', ADVANCEMENT_SETS, advancement_set)

    def dimensions(self, dimension: str) -> dict[str, dict]:
        """Returns the defined dimensions. You can add to this dict to add advancements to the datapack. """
        return self._get_json('dimension', None, dimension)

    def dimension_types(self, dimension_type: str) -> dict[str, dict]:
        """Returns defined dimension_types. You can add to this dict to add advancements to the datapack. """
        return self._get_json('dimension_type', None, dimension_type)

    def loot_tables(self, loot_table: str) -> dict[str, dict]:
        """Returns defined loot_tables. You can add to this dict to add advancements to the datapack. """
        return self._get_json('loot_tables', None, loot_table)

    def predicates(self) -> dict[str, dict]:
        """Returns defined predicates. You can add to this dict to add advancements to the datapack. """
        return self.json_directory('predicates')

    def recipes(self) -> dict[str, dict]:
        """Returns defined recipes. You can add to this dict to add advancements to the datapack. """
        return self.json_directory('recipes')

    def structures(self, structure: str) -> dict[str, dict]:
        """Returns defined structures. You can add to this dict to add advancements to the datapack. """
        return self._get_json('structures', None, structure)

    def worldgen(self, worldgen_set: str) -> dict[str, dict]:
        """Returns defined worldgen settings. You can add to this dict to add worldgen settings to the datapack. """
        return self._get_json('worldgen', WORLDGEN_SETS, worldgen_set)

    def json_directory(self, name: str) -> dict[str, dict]:
        """Returns an arbitrary directory at the top level of the data pack's JSON directory."""
        return self._get_json(name)

    def _get_json(self, directory, valid=None, which=None) -> dict:
        if valid:
            _in_group(valid, which)
        if not directory.endswith('/'):
            directory += '/'
        if not which.endswith('/'):
            which += '/'
        if directory not in self._json:
            self._json[directory] = {}
        if valid is None:
            return self._json[directory]
        return self._json[directory].setdefault(which, {})


class FunctionSet:
    """A set of functions. This corresponds to the top-level ``functions`` field in the datapack, and any of
    its child subdirectories. It enforces this maximum-two-level structure.

    The defined functions are available via the ``functions`` property. Typically, you will add functions via
    the add() method, which protects against duplicate functions. You can add, replace, or remove functions
    using the returned dict as well.
    """

    class _Functions(UserDict):
        """Dict that sets the added functions' parent."""

        def __init__(self, functions: FunctionSet):
            super().__init__()
            self.functions = functions

        def __setitem__(self, key: str, value: Function):
            ret = super().__setitem__(key, value)
            if not isinstance(value, Function):
                raise ValueError(f'{value.name}: Not a Function')
            value.parent = self.functions
            return ret

    def __init__(self, name: str, pack_or_parent: DataPack | FunctionSet = None):
        """Creates a function set.

        :param name: The set name. For the top level directory this is notional. For subdirectories, it is the
        directory name. The parent can be either a DataPack (for the top-level function directory) or that pack's
        FunctionSet.
        :param pack_or_parent: The parent of this set.
        """
        self.name = as_name(name)
        if isinstance(pack_or_parent, FunctionSet):
            # self.pack = pack_or_parent.pack
            self.parent = pack_or_parent
            self.parent.add_child(self)
            if self.parent.parent:
                raise ValueError(f'Only two levels of groups (sigh): {pack_or_parent.name} has a parent')
        elif isinstance(pack_or_parent, DataPack):
            self._pack = pack_or_parent
            self.parent = None
        else:
            self._pack = None
            self.parent = None

        self._functions = FunctionSet._Functions(self)
        self._kids: list[FunctionSet] = []

    @property
    def pack(self):
        if self._pack:
            return self._pack
        elif self.parent:
            return self.parent.pack
        return None

    @pack.setter
    def pack(self, pack: DataPack | None):
        self._pack = pack

    @property
    def full_name(self):
        if self.parent:
            parent_name = self.parent.full_name
            if not parent_name.endswith(':'):
                parent_name += '/'
            return f'{parent_name}{self.name}'
        if self.pack:
            return f'{self.pack.name}:'
        return self.name

    @classmethod
    def load(cls, path: Path | str, pack_or_parent: DataPack | FunctionSet = None, name=None) -> FunctionSet:
        """Loads the functions from the directory. Files that do not end in .mcfunction are skipped with a warning.
        If there are subdirectories, they will be loaded as well."""
        if not isinstance(path, Path):
            path = Path(path)
        if name is None:
            name = path.name
        elif path.name != name:
            path /= name
        fs = FunctionSet(name, pack_or_parent)
        if not path.exists():
            return fs
        for root, dirs, files in os.walk(path):
            for f in files:
                if not f.endswith(Function.SUFFIX):
                    print(f'Warning: {f} ignored: suffix required: {Function.SUFFIX}')
                    continue
                func = Function.load(Path(root) / f)
                fs.add(func)
            for d in dirs:
                FunctionSet.load(path / d, fs)
            # We're doing our own recursion, so just look at the top level
            break
        return fs

    def save(self, path: Path):
        """Saves the functions to the directory."""
        full_dir = self.path(path)
        if full_dir.exists():
            shutil.rmtree(full_dir)
        full_dir.mkdir(parents=True, exist_ok=True)
        for func in self.functions.values():
            func.save(full_dir / func.name)
        for child in self._kids:
            child.save(path)

    @property
    def functions(self) -> Mapping[str, Function]:
        """The functions in this set"""
        return self._functions

    @property
    def children(self) -> tuple[FunctionSet]:
        """The child FunctionSets of this set. This is read-only"""
        return tuple(self._kids)

    def path(self, path) -> Path:
        """The path to this function set."""
        return path / self._path_for()

    def _path_for(self) -> Path:
        if self.parent:
            return Path(self.parent.name) / self.name
        return Path(self.name)

    def add(self, function: Function) -> Function:
        """Adds a function to this set."""
        if function.name in self._functions:
            raise ValueError(f'{function.name}: duplicate function in {self.name}')
        self._functions[function.name] = function
        return function

    def child(self, name: str) -> FunctionSet | None:
        """Returns the named child FunctionSet of this function set."""
        for fs in self._kids:
            if fs.name == name:
                return fs
        return None

    def add_child(self, child: FunctionSet):
        """Adds a child FunctionSet."""
        self._kids.append(child)
        child.parent = self



"""Contains various data about vanilla Minecraft."""
from __future__ import annotations

import re
from collections import UserDict
from enum import Enum
from importlib.resources import files

from ._utils.fetch_things import ItemFetcher
from .base import COLORS, Nbt, NbtDef, to_id, to_name
from .commands import Block, Entity
from .enums import PotterySherd
from .simpler import Item, as_color_num

blocks: dict[str, Block] = {}
"""All blocks by name. See ``block_items`` if you want an item for a block."""
blocks_by_id: dict[str, Block] = {}
"""All blocks by ID."""
items: dict[str, Item] = {}
"""All items by name. Does not include items for blocks, but see ``block_items``."""
items_by_id: dict[str, Item] = {}
"""All items by ID."""
mobs: dict[str, Entity] = {}
mobs_by_id: dict[str, Entity] = {}

must_give_items: dict[str, Item] = {}
"""Items that are not in the creative inventory, by name."""
must_give_items_by_id: dict[str, Item] = {}
"""Items that are not in the creative inventory, by ID."""
operator_menu: dict[str, Item] = {}
"""Items that are only in the creative inventory if the 'Operator Menu' option is turned on."""


def __read_things(which: str, ctor):
    all_things = {}
    with (files(__package__).joinpath(f'all_{which}.txt')).open() as fp:
        for name in fp.readlines():
            name = name.strip()
            if not name or name[0] == '#':
                continue
            try:
                desc, id = re.split(r'\s*/\s*', name)
            except ValueError:
                id = desc = name
            thing = ctor(to_id(id), name=desc)
            all_things[desc] = thing

    return all_things, dict((t.id, t) for t in sorted(all_things.values(), key=lambda t: t.id))


def __read_lists():
    global blocks, blocks_by_id, items, items_by_id, must_give_items, must_give_items_by_id, mobs, mobs_by_id, \
        operator_menu
    blocks, blocks_by_id = __read_things('blocks', Block)
    items, items_by_id = __read_things('items', Block)
    mobs, mobs_by_id = __read_things('mobs', Entity)

    for item_name in ItemFetcher.must_give:
        item = must_give_items[item_name] = items[item_name]
        must_give_items_by_id[item.id] = item
        if item_name in ItemFetcher.operator_menu:
            operator_menu[item_name] = items[item_name]


__read_lists()


class _ItemForBlockDict(UserDict):
    """
    Dict that will return an Item(foo) for any foo that is a block, but does _not_ consider these in the dict. This
    lets the user say ``block_items['Oak Planks']`` and get a value, but also to ask if there is a special item for
    oak planks by testing ``'Oak Planks' in block_items``, which will be False. This works for the block names and
    block IDs.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        bi = args[0]
        for k, v in bi.items():
            try:
                block = blocks[k]
                self[block.id] = v
            except KeyError:
                pass

    def __getitem__(self, item):
        if item not in self:
            if item in blocks:
                return Item(item)


block_items = _ItemForBlockDict({
    'Bamboo Shoot': Item('Bamboo'),
    'Beetroots': Item('Beetroot Seeds'),
    'Carrots': Item('Carrot'),
    'Cave Vines': Item('Glow Berries'),
    'Cocoa': Item('Cocoa Beans'),
    'Fire': Item('Campfire'),
    'Lava': Item('Lava Bucket'),
    'Melon Stem': Item('Melon Seeds'),
    'Pumpkin Stem': Item('Pumpkin Seeds'),
    'Potatoes': Item('Potato'),
    'Powder Snow': Item('Powder Snow Bucket'),
    'Redstone Wire': Item('Redstone'),
    'Soul Fire': Item('Soul Campfire'),
    'Sweet Berry Bush': Item('Sweet Berries'),
    'Tall Seagrass': Item('Seagrass'),
    'Tripwire': Item('String'),
    'Water': Item('Water Bucket'),
})
"""Items for each kind of block. By default, this is simply ``Item(key)``, but there are some special cases where
there is no item for a block. This map takes a reasonable guess at the most reasonable item."""


class Color:
    """Represents a Minecraft color."""
    _num = 0

    def __init__(self, name: str, leather: int):
        self.name = name
        self.id = to_id(name)
        self.leather = leather
        self.num = Color._num
        Color._num += 1

    def __str__(self):
        return self.name


colors = (
    Color('White', 0xf9fffe),
    Color('Orange', 0xf9801d),
    Color('Magenta', 0xc74ebd),
    Color('Light Blue', 0x3ab3da),
    Color('Yellow', 0xfed83d),
    Color('Lime', 0x80c71f),
    Color('Pink', 0xf38baa),
    Color('Gray', 0x474f52),
    Color('Light Gray', 0x9d9d97),
    Color('Cyan', 0x169c9c),
    Color('Purple', 0x8932b8),
    Color('Blue', 0x3c44aa),
    Color('Brown', 0x835432),
    Color('Green', 0x5e7c16),
    Color('Red', 0xb02e26),
    Color('Black', 0x1d1d21),
)
"""The standard colors."""


class Instrument:
    """Data about a note block instrument."""

    def __init__(self, id, name, exemplar):
        """
        Creates a new instrument.

        :param id: The ID used in the note block's NBT
        :param name: The human-friendly name for the instrument.
        :param exemplar: One block you can put under the note block to get this instrument. Some instruments have
            several.
        """
        self.id = id
        self.name = name
        self.exemplar = exemplar


instruments = (
    Instrument('hat', 'High Hat', Block('Glass')),
    Instrument('basedrum', 'Base Drum', Block('Stone')),
    Instrument('snare', 'Snare Drum', Block('Sand')),
    Instrument('xylophone', 'Xylophone', Block('Bone Block')),
    Instrument('chime', 'Chime', Block('Packed Ice')),
    Instrument('harp', 'Harp', Block('grass_block', name='Other')),
    Instrument('guitar', 'Guitar', Block('white_wool', name='Wool')),
    Instrument('bass', 'Bass', Block('oak_planks', name='Wood')),
    Instrument('flute', 'Flute', Block('Clay')),
    Instrument('bell', 'Bell', Block('Gold Block')),
    Instrument('iron_xylophone', 'Iron Xylophone', Block('Iron Block')),
    Instrument('cow_bell', 'Cow Bell', Block('Soul Sand')),
    Instrument('didgeridoo', 'Didgeridoo', Block('Pumpkin')),
    Instrument('bit', 'Bit', Block('Emerald Block')),
    Instrument('banjo', 'Banjo', Block('Hay Block')),
    Instrument('pling', 'Pling', Block('Glowstone')),
)
"""The instruments note blocks can play."""


class Horse(Entity):
    """Data about a horse."""

    class Color(Enum):
        WHITE = 0
        CREAMY = 1
        CHESTNUT = 2
        BROWN = 3
        BLACK = 4
        GRAY = 5
        DARK_BROWN = 6

        def __int__(self):
            return self.value

    class Markings(Enum):
        NONE = 0 * 256
        WHITE = 1 * 256
        WHITE_FIELD = 2 * 256
        WHITE_SPOTS = 3 * 256
        BLACK_DOTS = 4 * 256

        def __int__(self):
            return self.value

    # Variants

    def __init__(self, name: str = None, color: int | Color = Color.WHITE, markings: int | Markings = Markings.NONE,
                 nbt: NbtDef = None):
        self.color = Horse.Color(color)
        self.markings = Horse.Markings(markings)
        self.variant = int(color) + int(markings)
        super().__init__('horse', name=name, nbt={'Variant': self.variant})
        if nbt:
            self.merge_nbt(nbt)
        self.tag_name = f'{to_id(name)}_horses'


horses = (
    Horse('White', Horse.Color.WHITE),
    Horse('Creamy', Horse.Color.CREAMY),
    Horse('Chestnut', Horse.Color.CHESTNUT),
    Horse('Brown', Horse.Color.BROWN),
    Horse('Black', Horse.Color.BLACK),
    Horse('Gray', Horse.Color.GRAY),
    Horse('Dark Brown', Horse.Color.DARK_BROWN),
)
"""The non-horse horses."""
other_horses = (
    Entity('Mule'),
    Entity('Donkey'),
    Entity('Skeleton Horse'),
    Entity('Zombie Horse'),
)
"""The non-horse horses."""
woods = ('Acacia', 'Bamboo', 'Birch', 'Cherry', 'Jungle', 'Mangrove', 'Oak', 'Dark Oak', 'Spruce')
"""The kinds of wood."""
stems = ('Warped', 'Crimson')
"""The kinds of stems."""
corals = ('Horn', 'Tube', 'Fire', 'Bubble', 'Brain')
"""The kinds of coral."""
tulips = ('Red', 'Orange', 'Pink', 'White')
"""The colors of tulips."""
small_flowers = ('Allium', 'Azure Bluet', 'Blue Orchid', 'Dandelion', 'Oxeye Daisy', 'Poppy')
"""The small flowers."""

moon_phases = (
    (206000, 'Full'),
    (38000, 'Waning Gibbous'),
    (62000, 'Three Quarters'),
    (86000, 'Waning Crescent'),
    (110000, 'New'),
    (134000, 'Waxing Crescent'),
    (158000, 'First Quarter'),
    (182000, 'Waxing Gibbous'),
)
"""The phases of the moon and the time for each."""

axolotls = ('Lucy', 'Wild', 'Gold', 'Cyan', 'Blue')
"""The kinds of axolotls."""

# noinspection SpellCheckingInspection
music_discs = (
    'music_disc_13',
    'music_disc_cat',
    'music_disc_blocks',
    'music_disc_chirp',
    'music_disc_far',
    'music_disc_mall',
    'music_disc_mellohi',
    'music_disc_stal',
    'music_disc_strad',
    'music_disc_ward',
    'music_disc_11',
    'music_disc_wait',
    'music_disc_otherside',
    'music_disc_pigstep',
    'music_disc_5',
)
"""The music discs."""


class Fish(Entity):
    """Data about a tropical fish."""
    _kinds = []

    def __init__(self, variant: int, name: str = None):
        """
        Creates a new fish with the given variant, and (if present) name. If not given, the desc() will be used as
        the name.
        """
        super().__init__('tropical_fish', name=name, nbt=Nbt(Variant=variant))
        self.variant = variant
        self.name = name if name else self.desc()

    @classmethod
    def _fish_kinds(cls):
        if not cls._kinds:
            cls._kinds = tuple(tropical_fish.keys())
        return cls._kinds

    @classmethod
    def variant(cls, kind: int | str, body_color: int | str, pattern_color: int | str) -> int:
        """
        Returns a variant for the described fish, to be used as a parameter to the Fish constructor.
        :param kind: The kind of fish, as either a name of a key in the ``tropical_fish`` dict, or an index of a key.
        :param body_color: The body color, as either a name or an index in the COLORS tuple.
        :param pattern_color: The pattern color, as either a name or an index in the COLORS tuple.
        """
        kind = Fish._to_kind(kind)

        return cls.spec_variant(kind < 6, kind % 6, body_color, pattern_color)

    @classmethod
    def spec_variant(cls, small: bool, which_body: int, body_color: int | str, pattern_color: int | str) -> int:
        """
        Returns a variant for the described fish, to be used as a parameter to the Fish constructor.
        :param small: Whether the fish is small.
        :param which_body: Which body type in the size, as a number from 0-5.
        :param body_color: The body color, as either a name or an index in the COLORS tuple.
        :param pattern_color: The pattern color, as either a name or an index in the COLORS tuple.
        """

        return int(small) | which_body << 8 | as_color_num(body_color) << 16 | as_color_num(pattern_color) << 24

    @classmethod
    def _to_kind(cls, kind):
        if isinstance(kind, str):
            kind = cls._fish_kinds().index(kind)
        return kind

    def is_small(self) -> bool:
        """Return whether this fish is small."""
        return self.variant & 0xff != 0

    def which_body(self) -> int:
        """Return which of the body types is used for this fish, in the range (0-5]."""
        return (self.variant & 0xff00) >> 8

    def kind(self) -> str:
        """Returns the name of the kind of fish."""
        return self._fish_kinds()[self.kind_num()]

    def kind_num(self) -> int:
        """Returns the number of this kind of fish. This is an index into ``tropical_fish.keys()``."""
        if self.is_small():
            return self.which_body()
        else:
            return self.which_body() + 6

    def body_color(self) -> str:
        """Returns the body color of the fish."""
        return COLORS[self.body_color_num()]

    def body_color_num(self) -> int:
        """Returns the number of the body color of the fish as an index into ``COLORS``."""
        return (self.variant & 0xff0000) >> 16

    def pattern_color(self) -> str:
        """Returns the pattern color of the fish."""
        return COLORS[self.pattern_color_num()]

    def pattern_color_num(self) -> int:
        """Returns the number of the pattern color of the fish as an index into ``COLORS``."""
        return (self.variant & 0xff000000) >> 24

    def desc(self) -> str:
        """
        Returns the description of this fish, which includes its body and pattern color (or just one if they are the
        same) and the body kind name.
        """
        body_color = to_name(self.body_color())
        pattern_color = to_name(self.pattern_color())
        if body_color == pattern_color:
            return f'{body_color} {self.kind()}'
        else:
            return f'{body_color}-{pattern_color} {self.kind()}'


tropical_fish = {
    'Spotty': (
        Fish(67110144, 'Goatfish'),
        Fish(50726144, 'Cotton Candy Betta')),
    'Brinely': (
        Fish(50660352, 'Queen Angelfish'),),
    'Dasher': (
        Fish(67699456, 'Yellowtail Parrotfish'),
        Fish(101253888, 'Parrotfish')),
    'Snooper': (
        Fish(235340288, 'Red Lipped Blenny'),),
    'Sunstreak': (
        Fish(50790656, 'Triggerfish'),
        Fish(118161664, 'Cichlid')),
    'Kob': (
        Fish(917504, 'Tomato Clownfish'),
        Fish(65536, 'Clownfish')),
    'Clayfish': (
        Fish(234882305, 'Emperor Red Snapper'),
        Fish(117441793, 'Butterflyfish'),
        Fish(16778497, 'Ornate Butterflyfish')),
    'Betty': (
        Fish(918529, 'Red Cichlid'),),
    'Blockfish': (
        Fish(67764993, 'Dottyback'),
        Fish(918273, 'Red Snapper')),
    'Glitter': (
        Fish(117441025, 'Moorish Idol'),),
    'Stripey': (
        Fish(117506305, 'Anemone'),),
    'Flopper': (
        Fish(117899265, 'Black Tang'),
        Fish(185008129, 'Blue Tang'),
        Fish(67371009, 'Yellow Tang'),
        Fish(67108865, 'Threadfin')),
}
"""The data for the predefined naturally-occurring tropical fish."""

trim_materials = sorted((
    'emerald', 'redstone', 'lapis', 'amethyst', 'quartz', 'netherite', 'diamond', 'gold', 'iron', 'copper'))

trim_patterns = sorted(
    ('coast', 'dune', 'eye', 'host', 'raiser', 'rib', 'sentry', 'shaper', 'silence', 'snout', 'spire', 'tide', 'vex',
     'ward', 'wayfinder', 'wild'))

armors = ('leather', 'chainmail', 'iron', 'golden', 'diamond', 'netherite')

sherds = tuple(x.value for x in PotterySherd)



import unittest

from pynecraft.info import Fish, block_items, blocks, blocks_by_id, items, items_by_id, mobs, mobs_by_id, \
    must_give_items, must_give_items_by_id
from pynecraft.simpler import Painting


class TestInfo(unittest.TestCase):
    def test_fish(self):
        var1 = Fish.variant('Spotty', 'Blue', 'RED')
        self.assertEqual(0x0e_0b_00_01, var1)
        fish = Fish(var1, 'foo')
        self.assertTrue(fish.is_small())
        self.assertEqual(0, fish.which_body())
        self.assertEqual('Spotty', fish.kind())
        self.assertEqual('blue', fish.body_color())
        self.assertEqual('red', fish.pattern_color())
        self.assertEqual('Blue-Red Spotty', fish.desc())
        var2 = Fish.variant('Glitter', 'GREEN', 'light blue')
        self.assertEqual(0x03_0d_03_00, var2)
        fish = Fish(var2, 'foo')
        self.assertFalse(fish.is_small())
        self.assertEqual(3, fish.which_body())
        self.assertEqual('Glitter', fish.kind())
        self.assertEqual('green', fish.body_color())
        self.assertEqual('light_blue', fish.pattern_color())
        self.assertEqual('Green-Light Blue Glitter', fish.desc())

    def test_full_lists(self):
        self.assertIn('Oak Sign', blocks)
        self.assertIn('oak_sign', blocks_by_id)
        self.assertNotIn('oak_wall_sign', blocks)

        self.assertIn('Book and Quill', items)
        self.assertIn('writable_book', items_by_id)
        self.assertNotIn('Oak Sign', items)

        self.assertIn('Cocoa', block_items)
        self.assertNotIn('Oak Sign', block_items)
        self.assertEqual('oak_sign', block_items['Oak Sign'].id)

        self.assertIn('Zombie', mobs)
        self.assertIn('zombie', mobs_by_id)

        self.maxDiff = None
        self.assertSequenceEqual(list(blocks.values()), sorted(blocks.values()))
        self.assertSequenceEqual(tuple(items.values()), sorted(items.values()))
        self.assertSequenceEqual(tuple(mobs.values()), sorted(mobs.values()))
        self.assertSequenceEqual(tuple(blocks_by_id.values()), sorted(blocks.values(), key=lambda t: t.id))
        self.assertSequenceEqual(tuple(items_by_id.values()), tuple(sorted(items.values(), key=lambda t: t.id)))
        self.assertSequenceEqual(tuple(mobs_by_id.values()), tuple(sorted(mobs.values(), key=lambda t: t.id)))

        for key, value in must_give_items.items():
            self.assertEqual(items[key], value)
        for key, value in must_give_items_by_id.items():
            self.assertEqual(items_by_id[key], value)

    def test_painting(self):
        self.assertEqual('bust', Painting('bust').nbt['variant'])
